#include "VisMonitoring/DQMServer/interface/VisDQMImageServer.h"
#include "VisMonitoring/DQMServer/interface/VisDQMRenderDaemon.h"
#include "classlib/utils/StringOps.h"
#include "classlib/utils/StringList.h"
#include "TObject.h"

using namespace lat;

class VisDQMLiveImageServer : public DQMImplNet<VisDQMObject>,
			      public VisDQMImageServer
{
public:
  static const uint32_t DQM_MSG_GET_IMAGE_LIVE	= 4;
  static const uint32_t DQM_REPLY_IMAGE_LIVE	= 105;

  VisDQMLiveImageServer(const ServerInfo &info, const std::string &dqmhost, int dqmport)
    : DQMImplNet<VisDQMObject>(info.argv[0]),
      VisDQMImageServer(info)
    {
      logme ()
	<< "INFO: DQM online backend started"
	<< ", bridging in path '" << info_.statedir
	<< "' and at request port " << info_.port
	<< " for data from " << dqmhost << ":" << dqmport
	<< ", black list in " << info_.blacklistFile << std::endl;

      // Establish the server side.
      debug(info_.debug);
      delay(100);
      sendScalarAsText(true);
      startLocalServer(info_.port);
      listenToCollector(dqmhost, dqmport);
    }

protected:
  // Return the logging prefix.
  virtual std::ostream &
  logme(void)
    {
      return DQMNet::logme();
    }

  // Check if the server should stop.
  virtual bool
  shouldStop(void)
    {
      return info_.stopfile.exists();
    }
  
  // Purge dead objects.  Clears object images.
  virtual void
  purgeDeadObject(VisDQMObject &o)
    {
      o.images.clear();
    }
  
  // When an object is released from wait, render the object image and
  // send back to the peer.
  virtual void
  releaseFromWait(Bucket *msg, WaitObject &w, Object *o)
    {
      // If imagespec is empty, the client wants just the object.
      if (w.info.empty())
	DQMNet::releaseFromWait(msg, w, o);

      // Otherwise, if we have an object, render it.
      else if (o)
	sendImageToPeer(msg, w.peer, 0,
			*static_cast<VisDQMObject *>(o),
			w.info, true);

      // Otherwise send something dummy.
      else
      {
	VisDQMObject dummy;
	dummy.version = 0;
	dummy.name = w.name;
	dummy.object = 0;
	dummy.reference = 0;
	dummy.flags = DQM_FLAG_DEAD;
	sendImageToPeer(msg, w.peer, 0, dummy, w.info, true);
      }
    }

  // Web server facing message handling.  Respond to object and image
  // requests.
  virtual bool
  onMessage(Bucket *msg, Peer *p, unsigned char *data, size_t len)
    {
      // Decode and process this message.
      uint32_t type;
      memcpy (&type, data + sizeof(uint32_t), sizeof (type));
      if (type == DQM_MSG_GET_IMAGE_LIVE)
      {
	if (debug_)
	  logme()
	    << "DEBUG: received message 'GET IMAGE LIVE' from peer "
	    << p->peeraddr << std::endl;

	if (len < 4*sizeof(uint32_t))
	{
	  logme()
	    << "ERROR: corrupt 'GET IMAGE LIVE' message of length " << len
	    << " from peer " << p->peeraddr << std::endl;
	  return false;
	}

	uint32_t namelen;
	uint32_t speclen;
	memcpy(&namelen, data + 2*sizeof(uint32_t), sizeof(namelen));
	memcpy(&speclen, data + 3*sizeof(uint32_t), sizeof(speclen));
	if (len != 4*sizeof(uint32_t) + namelen + speclen)
	{
	  logme()
	    << "ERROR: corrupt 'GET IMAGE LIVE' message of length " << len
	    << " from peer " << p->peeraddr
	    << ", expected length " << (4*sizeof(uint32_t))
	    << " + " << namelen << " + " << speclen << std::endl;
	  return false;
	}

	std::string name((char *) data + 4*sizeof(uint32_t), namelen);
	std::string spec((char *) data + 4*sizeof(uint32_t) + namelen, speclen);

	Peer *owner = 0;
	VisDQMObject *o = static_cast<VisDQMObject *>(findObject(0, name, &owner));
	if (! o)
	{
	  VisDQMObject dummy;
	  dummy.version = 0;
	  dummy.name = name;
	  dummy.object = 0;
	  dummy.reference = 0;
	  dummy.flags = DQM_FLAG_DEAD;
	  sendImageToPeer(msg, p, 0, dummy, spec, true);
	}
	else
	  sendImageToPeer(msg, p, owner, *o, spec, false);

	return true;
      }

      return DQMNet::onMessage(msg, p, data, len);
    }

  // Send an image of an object to a peer.
  void
  sendImageToPeer(Bucket *msg, Peer *p, Peer *owner, VisDQMObject &o,
		  const std::string &spec, bool force)
    {
      uint32_t datalen = 0;
      unsigned char *data = 0;

      // Mark the object referenced.
      o.lastreq = Time::current();

      // Validate the image request.
      const char *error = 0;
      VisDQMImgInfo *img = parseImageSpec(o, spec, error);
      if (! img)
      {
	logme()
	  << "ERROR: invalid image specification '"
	  << spec << "' for object '" << o.name
	  << "': " << error << std::endl;
	return;
      }

      // The object is known and the image specification is valid.  Check
      // if we have an already rendered image, and if so, use it as such.
      if (img->version == o.version)
      {
	datalen = img->imgdata.size();
	data = (datalen ? &img->imgdata[0] : 0);
      }

      // Otherwise check if we have data for the object.  If not, and we
      // are not forced to render an image, delay until the data arrives.
      else if (o.rawdata.empty() && !force)
	return waitForData(p, o.name, spec, owner);

      // We have an object, but no valid image yet.  Render it now.  This
      // may require us to reconstruct the object from the raw data.
      else
      {
	// Try extracting the object and its quality reports if we haven't
	// yet done so.  If we fail even partially, we nuke the lot.  Note
	// that whenever the object is updated, the o.object is destroyed.
	if (! o.object && o.rawdata.size() && ! reconstructObject(o) && ! force)
	  logme()
	    << "ERROR: failed to reconstruct object '"
	    << o.name << "' from server data\n";

	doRender(o, *img);
	datalen = img->imgdata.size();
	data = (datalen ? &img->imgdata[0] : 0);
      }

      // Build the response.
      uint32_t words [5];
      uint32_t namelen = o.name.size();
      uint32_t speclen = spec.size();
      words[0] = sizeof(words) + namelen + speclen + datalen;
      words[1] = DQM_REPLY_IMAGE_LIVE;
      words[2] = namelen;
      words[3] = speclen;
      words[4] = datalen;

      msg->data.reserve(msg->data.size() + words[0]);
      copydata(msg, &words[0], sizeof(words));
      if (namelen)
	copydata(msg, &o.name[0], namelen);
      if (speclen)
	copydata(msg, &spec[0], speclen);
      if (datalen)
	copydata(msg, data, datalen);
    }
};

class VisDQMLiveDaemon : public VisDQMRenderDaemon
{
  std::string	dqmhost_;
  int		dqmport_;

public:
  VisDQMLiveDaemon(int argc, char **argv)
    : VisDQMRenderDaemon(argc, argv),
      dqmhost_("localhost"),
      dqmport_(9091)
    {}

protected:
  virtual bool
  parseOption(int &i, int argc, char **argv)
    {
      if (i < argc-1 && ! strcmp(argv[i], "--collector"))
      {
	StringList server = StringOps::split(argv[++i], ':', 0, 2, 0, -1);
	if (server[0].size ()) dqmhost_ = server[0];
	if (server.size () == 2) dqmport_ = atoi(server[1].c_str());
	return true;
      }
      return false;
    }

  virtual std::string
  showOptions(void)
    {
      return " [--collector HOST[:PORT]]";
    }

  virtual void
  runServer(const VisDQMImageServer::ServerInfo &info)
    {
      VisDQMLiveImageServer server(info, dqmhost_, dqmport_);
      server.run();
    }
};


int main(int argc, char **argv)
{
  VisDQMLiveDaemon d(argc, argv);
  return d.execute();
}
