// Shut up warning about _POSIX_C_SOURCE mismatch, caused by a
// harmless conflict between _GNU_SOURCE and python header files.
#define NDEBUG 1
#include <features.h>
#if _POSIX_C_SOURCE != 200112L
# undef _POSIX_C_SOURCE
# define _POSIX_C_SOURCE 200112L
#endif

#include "VisMonitoring/DQMServer/interface/Objects.h"
#include "VisMonitoring/DQMServer/interface/NatSort.h"
#include "VisMonitoring/DQMServer/interface/StringAtom.h"
#include "DQMServices/Core/interface/DQMNet.h"
#include "FWCore/Utilities/interface/EDMException.h"

#include "classlib/utils/DebugAids.h"
#include "classlib/utils/Signal.h"
#include "classlib/utils/Regexp.h"
#include "classlib/utils/RegexpMatch.h"
#include "classlib/utils/StringOps.h"
#include "classlib/utils/StringFormat.h"
#include "classlib/iobase/Filename.h"
#include "classlib/iobase/IOStatus.h"
#include "classlib/iobase/File.h"

#include "boost/shared_ptr.hpp"
#include "boost/python.hpp"
#include "boost/python/stl_iterator.hpp"

#include "TObjString.h"
#include "TROOT.h"

#include "sqlite3.h"
#include <set>
#include <ext/hash_map>

using namespace lat;
using namespace boost;
namespace ext = __gnu_cxx;
namespace py = boost::python;
typedef std::set<std::string> StringSet;
typedef std::set<StringAtom> StringAtomSet;
typedef std::vector<StringAtom> StringAtomList;

static StringAtomTree stree(1024*1024);
static Regexp RX_THOUSANDS("(\\d)(\\d{3}($|\\D))");
static const uint32_t DQM_FLAG_HAS_REFERENCE = 0x04000000;
static const uint32_t HAS_ALARM = (DQMNet::DQM_FLAG_REPORT_ERROR
				   | DQMNet::DQM_FLAG_REPORT_WARNING
				   | DQMNet::DQM_FLAG_REPORT_OTHER);

namespace __gnu_cxx
{
  template<> struct hash<StringAtom>
  {
    size_t operator()(const StringAtom &s) const
    { return s.index(); }
  };

  template<> struct hash<std::string>
  {
    size_t operator()(const std::string &s) const
    { return __stl_hash_string(s.c_str()); }
  };
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class PyReleaseInterpreterLock
{
  PyThreadState *save_;
public:
  PyReleaseInterpreterLock(void) { save_ = PyEval_SaveThread(); }
  ~PyReleaseInterpreterLock(void) { PyEval_RestoreThread(save_); }
};

class Lock
{
  pthread_mutex_t *lock_;
public:
  Lock(pthread_mutex_t *lock) { pthread_mutex_lock(lock_ = lock); }
  ~Lock(void)                 { pthread_mutex_unlock(lock_); }
};

// Always abort on assertion failure.
static char
onAssertFail(const char *message)
{
  std::cout.flush();
  fflush(stdout);
  std::cerr.flush();
  fflush(stderr);
  std::cerr << message << "ABORTING\n";
  return 'a';
}

static std::ostream &
lograw(const char *label)
{
  return std::cout
    << Time::current().format(true, "[%d/%b/%Y:%H:%M:%S]")
    << label;
}

static std::ostream &
loginfo(void)
{
  return lograw("  INFO: ");
}

static std::ostream &
logerr(void)
{
  return lograw("  ERROR: ");
}

static std::ostream &
logwarn(void)
{
  return lograw("  WARNING: ");
}

static std::string
thousands(const std::string &arg)
{
  std::string result;
  std::string tmp(arg);
  result.reserve(arg.size() + arg.size()/3 + 1);

  while (true)
  {
    result = StringOps::replace(tmp, RX_THOUSANDS, "\\1'\\2");
    if (result == tmp)
      break;

    tmp = result;
  }

  return result;
}

std::string
stringToJSON(const std::string &x, bool emptyIsNone = false)
{
  if (emptyIsNone && x.empty())
    return "None";

  char buf[8];
  std::string result;
  result.reserve(2 * x.size());
  result += '"';
  for (size_t i = 0, e = x.size(); i < e; ++i)
  {
    unsigned int ch = (unsigned char) x[i];
    if (ch == '\\')
    {
      result += '\\';
      result += '\\';
    }
    else if (ch == '"')
    {
      result += '\\';
      result += (char) ch;
    }
    else if (ch == '\n')
    {
      result += '\\';
      result += 'n';
    }
    else if (ch == '\r')
    {
      result += '\\';
      result += 'r';
    }
    else if (ch == '\t')
    {
      result += '\\';
      result += 't';
    }
    else if (! isprint(ch))
    {
      sprintf(buf, "\\u%04x", ch);
      result += buf;
    }
    else
      result += (char) ch;
  }
  
  result += '"';
  return result;
}

static void
updateRunInfo(const char *data,
	      long &curRunNr,
	      long &curLumiNr,
	      long &curEventNr)
{
  char *endp = 0;
  long value;
  if (! strncmp(data, "<iRun>i=", 8)
      && (value = strtol(data + 8, &endp, 10)) > curRunNr
      && endp
      && ! strcmp(endp, "</iRun>"))
    curRunNr = value;
  else if (! strncmp(data, "<iLumiSection>i=", 16)
	   && (value = strtol(data + 16, &endp, 10)) > curLumiNr
	   && endp
	   && ! strcmp(endp, "</iLumiSection>"))
    curLumiNr = value;
  else if (! strncmp(data, "<iEvent>i=", 10)
	   && (value = strtol(data + 10, &endp, 10)) > curEventNr
	   && endp
	   && ! strcmp(endp, "</iEvent>"))
    curEventNr = value;
}

static shared_ptr<Regexp>
rx(const std::string &match, int options = 0)
{
  shared_ptr<Regexp> r;
  if (! match.empty())
  {
    r.reset(new Regexp(match, options));
    if (! r->valid())
      throw std::runtime_error(StringFormat
			       ("Invalid regular expression syntax at"
				" character %2 of '%1': %3")
			       .arg(match)
			       .arg(r->errorOffset())
			       .arg(r->errorMessage()));
    r->study();
  }
  return r;
}

static void
buildParentNames(StringAtomList &to, const StringAtom &from)
{
  const std::string &fromstr = from.string();
  std::string parentstr;
  parentstr.reserve(fromstr.size());
  to.reserve(std::count(fromstr.begin(), fromstr.end(), '/'));

  for (size_t slash = 0; slash < fromstr.size();
       slash = fromstr.find('/', ++slash))
  {
    parentstr.assign(fromstr, 0, slash);
    to.push_back(StringAtom(&stree, parentstr));
  }
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
struct VisDQMAxisOptions;
struct VisDQMDrawOptions;
struct VisDQMLayoutItem;
struct VisDQMLayoutRow;
struct VisDQMItem;
struct VisDQMStatus;
struct VisDQMShownItem;
struct VisDQMShownRow;
struct VisDQMStatus;
class VisDQMSource;

typedef std::vector< shared_ptr<VisDQMLayoutItem> >	      VisDQMLayoutItems;
typedef std::vector< shared_ptr<VisDQMLayoutRow> >	      VisDQMLayoutRows;
typedef std::vector< shared_ptr<VisDQMShownItem> >	      VisDQMShownItems;
typedef std::vector< shared_ptr<VisDQMShownRow> >	      VisDQMShownRows;
typedef ext::hash_map<StringAtom, shared_ptr<VisDQMItem> >    VisDQMItems;
typedef ext::hash_map<StringAtom, VisDQMStatus>		      VisDQMStatusMap;
typedef std::map<StringAtom, VisDQMDrawOptions>		      VisDQMDrawOptionMap;

struct VisDQMAxisOptions
{
  std::string			min;
  std::string			max;
  std::string			type;
};

struct VisDQMDrawOptions
{
  VisDQMAxisOptions		xaxis;
  VisDQMAxisOptions		yaxis;
  VisDQMAxisOptions		zaxis;
  std::string			drawopts;
  std::string			withref;
};

struct VisDQMLayoutItem
{
  StringAtom			path;
  std::string			desc;
  VisDQMDrawOptions		drawopts;
};

struct VisDQMLayoutRow
{
  VisDQMLayoutItems		columns;
};

struct VisDQMItem
{
  int32_t			fileid;
  uint32_t			flags;
  uint64_t			version;
  std::string			realname;
  StringAtom			name;
  StringAtomList		parents;
  DQMNet::DataBlob		data;
  VisDQMSource		 	*plotter;
  shared_ptr<VisDQMLayoutRows>	layout;
};

struct VisDQMShownItem
{
  uint64_t			version;
  int				nalarm;
  int				nlive;
  StringAtom			name;
  std::string			desc;
  VisDQMSource			*plotter;
  VisDQMDrawOptions		drawopts;
};

struct VisDQMShownRow
{
  VisDQMShownItems		columns;
};

struct VisDQMStatus
{
  int				isdir;
  int				nleaves;
  int				nalarm;
  int				nlive;
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
static void
translateDrawOptions(py::dict opts, VisDQMDrawOptions &o)
{
  py::stl_input_iterator<std::string> ki(opts), ke;
  for ( ; ki != ke; ++ki)
  {
    std::string key(*ki);
    std::string value(py::extract<std::string>(opts.get(key)));
    if (key == "withref")        o.withref = value;
    else if (key == "drawopts")  o.drawopts = value;
    else if (key == "xmin")      o.xaxis.min = value;
    else if (key == "xmax")      o.xaxis.max = value;
    else if (key == "xtype")     o.xaxis.type = value;
    else if (key == "ymin")      o.yaxis.min = value;
    else if (key == "ymax")      o.yaxis.max = value;
    else if (key == "ytype")     o.yaxis.type = value;
    else if (key == "zmin")      o.zaxis.min = value;
    else if (key == "zmax")      o.zaxis.max = value;
    else if (key == "ztype")     o.zaxis.type = value;
  }
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMSource
{
public:
  VisDQMSource(void)
    {}

  virtual ~VisDQMSource(void)
    {}

  virtual const char *
  plotter(void) const
    {
      return "unknown";
    }

  virtual void
  prepareSession(py::dict session)
    {}

  virtual void
  prescan(void)
    {}

  virtual void
  scan(VisDQMItems &result,
       StringSet &datasets,
       StringSet &steps,
       StringSet &runs,
       const std::string &dataset,
       const std::string &step,
       const std::string &runnr,
       long &curRunNr,
       long &curLumiNr,
       long &curEventNr,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm)
    {}

protected:
  static bool
  hasSessionKey(const py::dict &session, const char *key)
    {
      return py::extract<std::string>(session.get(key, ""))() != "";
    }

private:
  VisDQMSource(VisDQMSource &);
  VisDQMSource &operator=(VisDQMSource &);
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMUnknownSource : public VisDQMSource
{
public:
  VisDQMUnknownSource(void)
    {}
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMLayoutSource : public VisDQMSource
{
  struct FileData
  {
    Time		mtime;
    VisDQMItems		items;
  };

  typedef std::map<Filename, FileData> FileDataMap;
  typedef std::list<Filename> FileNameList;

  pthread_mutex_t	lock_;
  FileDataMap		data_;
  FileNameList		files_;

public:
  VisDQMLayoutSource(py::tuple filenames)
    {
      pthread_mutex_init(&lock_, 0);
      py::stl_input_iterator<std::string> begin(filenames), end;
      files_.insert(files_.end(), begin, end);
    }

  ~VisDQMLayoutSource(void)
    {}

  virtual void
  prescan(void)
    {
      Lock gate(&lock_);
      for (FileNameList::iterator i = files_.begin(), e = files_.end(); i != e; ++i)
      {
	IOStatus st;
	if (! st.fill(*i))
	  data_.erase(*i);
	else
	{
	  FileData &d = data_[*i];
	  if (d.mtime != st.m_mtime)
	  {
	    d.mtime = st.m_mtime;
	    d.items.clear();
	    try
	    {
	      py::dict items;
	      py::dict locals;
	      locals["dqmitems"] = items;
	      py::exec_file(i->name(), py::import("GuiDQM").attr("__dict__"), locals);

	      py::stl_input_iterator<std::string> ki(items), ke;
	      for ( ; ki != ke; ++ki)
	      {
		shared_ptr<VisDQMItem> i(new VisDQMItem);
		i->fileid = 0;
		i->flags = 0;
		i->version = st.m_mtime.ns();
		i->name = StringAtom(&stree, *ki);
		i->plotter = 0;
		buildParentNames(i->parents, i->name);
		d.items[i->name] = i;

		i->layout.reset(new VisDQMLayoutRows);
		py::object rows = py::extract<py::object>(items.get(*ki).attr("layout"));
		py::stl_input_iterator<py::object> ri(rows), re;
		for ( ; ri != re; ++ri)
		{
		  py::stl_input_iterator<py::object> ci(*ri), ce;
		  shared_ptr<VisDQMLayoutRow> row(new VisDQMLayoutRow);
		  for ( ; ci != ce; ++ci)
		  {
		    py::object pycol = *ci;
		    py::extract<std::string> pycolstr(pycol);
		    py::extract<py::dict> pycoldict(pycol);
		    shared_ptr<VisDQMLayoutItem> col(new VisDQMLayoutItem);
		    if (pycol.ptr() == Py_None)
		      // None, leave name as empty.
		      col->path = StringAtom(&stree, std::string());
		    else if (pycolstr.check())
		      // Basic string, use as a name.
		      col->path = StringAtom(&stree, pycolstr());
		    else if (pycoldict.check())
		    {
		      // Complete description, copy over.
		      py::dict colobj = pycoldict();
		      col->path = StringAtom(&stree, py::extract<std::string>(colobj.get("path")));
		      col->desc = py::extract<std::string>(colobj.get("description", ""));
		      if (colobj.has_key("draw"))
			translateDrawOptions(py::extract<py::dict>(colobj.get("draw")), col->drawopts);
		    }
		    else
		      logwarn()
			<< "layout element " << i->name.string()
			<< "[" << i->layout->size() << ", " << row->columns.size()
			<< "] is neither 'None', string nor a dict!\n";

		    row->columns.push_back(col);
		  }
		  
		  i->layout->push_back(row);
		}
	      }
	    }
	    catch (const py::error_already_set &e)
	    {
	      logerr() << "failed to update layouts from " << *i << std::endl;
	      PyErr_Print();
	    }
	  }
	}
      }
    }

  virtual void
  scan(VisDQMItems &result,
       StringSet &datasets,
       StringSet &steps,
       StringSet &runs,
       const std::string &dataset,
       const std::string &step,
       const std::string &runnr,
       long &curRunNr,
       long &curLumiNr,
       long &curEventNr,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm)
    {
      Lock gate(&lock_);
      FileDataMap::iterator fi, fe;
      VisDQMItems::iterator oi, oe;
      for (fi = data_.begin(), fe = data_.end(); fi != fe; ++fi)
      {
	FileData &fd = fi->second;
	result.resize(result.size() + fd.items.size());
	for (oi = fd.items.begin(), oe = fd.items.end(); oi != oe; ++oi)
	{
	  VisDQMItem &o = *oi->second;

	  if (rxmatch && ! rxmatch->match(o.name.string()))
	    continue;

	  if (rxsearch && rxsearch->search(o.name.string()) < 0)
	    continue;

	  if (alarm && ((o.flags & HAS_ALARM) != 0) != (*alarm == true))
	    continue;

	  result[o.name] = oi->second;
	}
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
/** A background thread assisting #VisDQMLiveSource.  Listens to
    changes in live DQM data in upstream sources and maintains a
    up-to-date summary relevant for the web service. */
class VisDQMLiveThread : public DQMBasicNet
{
  static const uint64_t DEAD_PRUNE = 900000000000ULL; // 15 minutes
  VisDQMSource	*owner_;
  bool		stop_;
  VisDQMItems	items_;

public:
  VisDQMLiveThread(VisDQMSource *owner, bool verbose, const std::string &host, int port)
    : DQMBasicNet("VisDQMLiveThread"),
      owner_(owner),
      stop_(false)
    {
      logme()
	<< "INFO: DQM live thread started"
	<< ", listening for data from "
	<< host << ":" << port << "\n";

      debug(verbose);
      delay(100);
      listenToCollector(host, port);
      start();
    }

  void
  stop(void)
    {
      lock();
      stop_ = true;
      unlock();
    }

  void
  scan(VisDQMItems &result,
       long &curRunNr, long &curLumiNr, long &curEventNr,
       Regexp *rxmatch, Regexp *rxsearch, bool *alarm)
    {
      lock();

      PeerMap::iterator pi, pe;
      ObjectMap::iterator oi, oe;
      VisDQMItems::iterator di, de;
      uint64_t now = Time::current().ns();

      // First mark all replica objects no longer known.
      for (di = items_.begin(), de = items_.end(); di != de; ++di)
	di->second->flags |= DQM_FLAG_ZOMBIE;

      // Scan still known objects, and update replica if necessary.
      for (pi = peers_.begin(), pe = peers_.end(); pi != pe; ++pi)
	for (oi = pi->second.objs.begin(), oe = pi->second.objs.end(); oi != oe; ++oi)
	{
	  Object &o = oi->second;
	  bool isnew = o.flags & DQM_FLAG_NEW;
	  StringAtom name(&stree, o.name);
	  shared_ptr<VisDQMItem> &i = items_[name];
	  if (! i)
	  {
	    i.reset(new VisDQMItem);
	    i->fileid = 0;
	    i->flags = 0;
	    i->version = 0;
	    i->name = name;
	    i->plotter = owner_;
	    buildParentNames(i->parents, i->name);
	    isnew = true;
	  }

	  i->flags &= ~DQM_FLAG_ZOMBIE;
	  if (isnew || o.flags != i->flags)
	  {
	    o.flags &= ~DQM_FLAG_NEW;
	    i->flags = o.flags;
	    i->version = o.version;
	    if (o.flags & DQM_FLAG_SCALAR)
	    {
	      i->data.clear();
	      extractScalarData(i->data, o);
	      i->data.push_back(0);
	    }
	  }
	}

      // Finally scan replicas again.  Remove too old dead objects as
      // we run into them.  Match remaining valid objects against the
      // caller's filters and add matching ones to the result set.
      result.resize(result.size() + items_.size());
      for (di = items_.begin(), de = items_.end(); di != de; )
      {
	VisDQMItem &i = *di->second;

	if (i.flags & DQM_FLAG_ZOMBIE)
	  i.flags |= DQM_FLAG_DEAD;

	if ((i.flags & DQM_FLAG_DEAD) && now - i.version > DEAD_PRUNE)
	{
	  items_.erase(di++);
	  continue;
	}

	if ((rxmatch && ! rxmatch->match(i.name.string()))
	    || (rxsearch && rxsearch->search(i.name.string()) < 0)
	    || (alarm && ((i.flags & HAS_ALARM) != 0) != (*alarm == true)))
	{
	  ++di;
	  continue;
	}

	if ((i.flags & DQM_FLAG_SCALAR)
	    && i.name.string().find("/EventInfo/i") != std::string::npos)
	  updateRunInfo((const char *) &i.data[0], curRunNr, curLumiNr, curEventNr);

	result[i.name] = di->second;
	++di;
      }
      unlock();
    }

protected:
  // Check if the server should stop.
  virtual bool
  shouldStop(void)
    {
      return stop_;
    }
};

class VisDQMLiveSource : public VisDQMSource
{
  std::string mydataset_;
  std::string mystep_;
  std::string myrun_;
  VisDQMLiveThread *thread_;

public:
  VisDQMLiveSource(py::object gui, py::dict opts)
    : mydataset_(py::extract<std::string>(opts["dataset"])),
      mystep_(py::extract<std::string>(opts["step"])),
      myrun_(py::extract<std::string>(opts["runnr"])),
      thread_(new VisDQMLiveThread(this,
				   py::extract<bool>(opts["verbose"]),
				   py::extract<std::string>(opts["dqmhost"]),
				   py::extract<int>(opts["dqmport"])))
    {}

  ~VisDQMLiveSource(void)
    {
      if (thread_)
	stop();
    }

  virtual const char *
  plotter(void) const
    {
      return "current";
    }

  std::string
  path(const std::string &dataset,
       const std::string &step,
       const std::string &runnr,
       const std::string &name)
    {
      if (dataset == mydataset_ && step == mystep_ && runnr == myrun_)
	return name;
      return "";
    }
  
  virtual void
  stop(void)
    {
      thread_->stop();
      delete thread_;
      thread_ = 0;
    }

  virtual void
  prepareSession(py::dict session)
    {
      if (! hasSessionKey(session, "dqm.dataset")
	  && ! hasSessionKey(session, "dqm.step")
	  && ! hasSessionKey(session, "dqm.runnr"))
      {
	session["dqm.dataset"] = mydataset_;
	session["dqm.step"] = mystep_;
	session["dqm.runnr"] = myrun_;
      }
    }

  virtual void
  scan(VisDQMItems &result,
       StringSet &datasets,
       StringSet &steps,
       StringSet &runs,
       const std::string &dataset,
       const std::string &step,
       const std::string &runnr,
       long &curRunNr,
       long &curLumiNr,
       long &curEventNr,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm)
    {
      datasets.insert(mydataset_);
      steps.insert(mystep_);
      runs.insert(myrun_);

      if (dataset == mydataset_ && step == mystep_ && runnr == myrun_)
	thread_->scan(result, curRunNr, curLumiNr, curEventNr,
		      rxmatch, rxsearch, alarm);
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMArchiveSource : public VisDQMSource
{
  /// Nanoseconds to keep data cached before purging it.
  static const int64_t MAX_CACHE_LIFE = 900000000000LL;

  struct RunData
  {
    Time		stamp;
    long		runId;
    long		datasetId;
    long		stepId;
    long		runNumber;
    long		lumiNumber;
    long		eventNumber;
    VisDQMItems		items;
  };

  typedef ext::hash_map<long, std::string> IdKeyMap;
  typedef ext::hash_map<std::string, long> KeyIdMap;
  typedef std::map<std::string, RunData>   RunCache;

  Filename		masterFileName_;
  Filename		localFileName_;
  Time			lastUpdate_;

  pthread_mutex_t	lock_;
  sqlite3		*db_;
  KeyIdMap		datasetToId_;
  KeyIdMap		stepToId_;
  IdKeyMap		idToDataset_;
  IdKeyMap		idToStep_;
  IdKeyMap		idToSystem_;
  IdKeyMap		idToCategory_;
  IdKeyMap		idToFile_;
  IdKeyMap		idToName_;

  RunCache		runCache_;

public:
  VisDQMArchiveSource(py::object gui, py::dict opts)
    {
      pthread_mutex_init(&lock_, 0);
      masterFileName_ = Filename(py::extract<std::string>(opts["master"]));
      localFileName_ = Filename(py::extract<std::string>(opts["local"]));
      lastUpdate_ = 0;
      db_ = 0;
    }

  virtual const char *
  plotter(void) const
    {
      return "archive";
    }

  py::tuple
  path(const std::string &dataset,
       const std::string &step,
       const std::string &runnr,
       const std::string &namestr)
    {
      std::string runkey;
      std::string fname;
      std::string realname;
      fillRunKey(runkey, dataset, step, runnr);
      
      StringAtom name(&stree, namestr, StringAtom::TestOnly);
      {
	Lock gate(&lock_);
	update(runkey);
	RunCache::iterator pos = runCache_.find(runkey);
	if (pos != runCache_.end())
	{
	  RunData &d = pos->second;
	  VisDQMItems::iterator opos = d.items.find(name);
	  if (opos != d.items.end())
	  {
	    VisDQMItem &o = *opos->second;
	    fname = idToFile_[o.fileid];
	    realname = o.realname;
	  }
	}
      }

      return py::make_tuple(fname, realname);
    }
  
  // Utility function to reset current data contents.
  void
  reset(void)
    {
      if (db_)
        dbclose(&db_);

      datasetToId_.clear();
      stepToId_.clear();
      idToDataset_.clear();
      idToStep_.clear();
      idToSystem_.clear();
      idToCategory_.clear();
      idToFile_.clear();
      idToName_.clear();
      runCache_.clear();
    }

  // Utility function to fetch a simple id <-> name table.
  void
  fillMap(const std::string &table, IdKeyMap &map)
    {
      int rc;
      char *errmsg = 0;
      if ((rc = sqlite3_exec(db_, ("select id, name from t_" + table).c_str(),
			     &setupIdMapping, &map, &errmsg)) != SQLITE_OK)
	throw dberror("sqlite3_exec", rc, errmsg);
    }

  static int
  setupIdMapping(void *ptr, int argc, char **argv, char ** /* columns */)
    {
      (*(IdKeyMap *) ptr)[atol(argv[0])] = argv[1];
      return 0;
    }

  static int
  setupRunData(void *ptr, int argc, char **argv, char ** /* columns */)
    {
      VisDQMArchiveSource *self = (VisDQMArchiveSource *) ptr;
      long runnr = atol(argv[0]);
      long dataset = atol(argv[1]);
      long step = atol(argv[2]);

      ASSERT(self->idToDataset_.count(dataset) > 0);
      ASSERT(self->idToStep_.count(step) > 0);

      std::string key = StringFormat("%1:%2:%3")
			.arg(runnr)
			.arg(self->idToDataset_[dataset])
			.arg(self->idToStep_[step]);

      RunData &d = self->runCache_[key];
      d.stamp = 0;
      d.runId = runnr;
      d.datasetId = dataset;
      d.stepId = step;
      d.runNumber = 0;
      d.lumiNumber = 0;
      d.eventNumber = 0;

      return 0;
    }

  static void
  dbclose(sqlite3 **db)
    {
#if 0
      while (sqlite3_stmt *stmt = sqlite3_next_stmt(*db, 0))
        sqlite3_finalize(stmt);
#endif

      int rc = sqlite3_close(*db);
      if (rc != SQLITE_OK)
        logwarn() << "sqlite3_close(): resource leak: failed to close"
		  << " the database, error code was " << rc << '\n';

      *db = 0;
    }

  static std::runtime_error
  dberror(const char *context, int rc, const char *message)
    {
      std::string msg = StringFormat("%1(): error: %2 (sqlite return code %3)")
			.arg(context)
			.arg(message)
			.arg(rc);
      sqlite3_free((void *) message);
      return std::runtime_error(msg);
    }

  static std::runtime_error
  dberror(const char *context,
	  int rc, bool close, sqlite3 **db,
	  const std::string &prefix)
    {
      std::runtime_error err
	(StringFormat("%1(): error: %2, %3 (database error code %4,"
		      " sqlite return code %5)")
	 .arg(context)
	 .arg(prefix)
	 .arg(sqlite3_errmsg(*db))
	 .arg(sqlite3_errcode(*db))
	 .arg(rc));

      if (close)
	dbclose(db);

      return err;
    }

  static std::runtime_error
  dberror(const char *context, int rc, sqlite3 *db, sqlite3_stmt *stmt)
    {
      std::runtime_error err
	(StringFormat("%1(): error: cannot execute statement, %2"
		      " (database error code %3, sqlite return code %4)")
	 .arg(context)
	 .arg(sqlite3_errmsg(db))
	 .arg(sqlite3_errcode(db))
	 .arg(rc));

      sqlite3_finalize(stmt);
      return err;
    }

  // Refresh data from the database for a run.  Checks if the master
  // database exists, and if so, if it has changed relative to the
  // last copy and the time we last read in the data.  If so, discards
  // any existing cached data and copies the file again.  Then checks
  // to see if the data for the requested run is already cached, and
  // if not, reads it in.  Also discards cached data for runs that
  // have not been accessed recently, where "recent" is MAX_CACHE_LIFE
  // seconds.
  void
  update(const std::string &runkey)
    {
      // First discard not recently accessed, still cached data.
      Time now = Time::current();
      RunCache::iterator i, e;
      for (i = runCache_.begin(), e = runCache_.end(); i != e; ++i)
	if (i->first != runkey
	    && i->second.stamp > 0
	    && now - i->second.stamp >= MAX_CACHE_LIFE)
	{
	  i->second.stamp = 0;
	  i->second.items.clear();
	}

      // See if we need to copy the file over again.
      IOStatus master;
      IOStatus local;
      bool newdata = false;
      bool havemaster = master.fill(masterFileName_);
      bool havelocal = local.fill(localFileName_);
      if (havemaster
	  && (! havelocal
	      || local.m_mtime != master.m_mtime
	      || local.m_size != master.m_size))
      {
	try
	{
	  reset();
	  loginfo()
	    << "copying db " << masterFileName_
	    << " to " << localFileName_ << std::endl;

	  IOSize nread;
	  std::vector<char> buf(64*1024, 0);
	  File localfile(localFileName_,
			 IOFlags::OpenWrite
			 | IOFlags::OpenCreate
			 | IOFlags::OpenTruncate);
	  File masterfile(masterFileName_);
	  while ((nread = masterfile.read(&buf[0], buf.size())) > 0)
	    localfile.write(&buf[0], nread);
	  localfile.close();
	  masterfile.close();
	  struct timeval t[2];
	  t[0].tv_sec = master.m_atime.ns() / 1000000000;
	  t[0].tv_usec = (master.m_atime.ns() % 1000000000) / 1000;
	  t[1].tv_sec = master.m_mtime.ns() / 1000000000;
	  t[1].tv_usec = (master.m_mtime.ns() % 1000000000) / 1000;
	  utimes(localFileName_, t);
	  local = master;
	  havelocal = true;
	  newdata = true;
	}
	catch (Error &e)
	{
	  logerr()
	    << "failed to copy database " << masterFileName_
	    << " to " << localFileName_
	    << ": " << e.explain()
	    << std::endl;
	}
      }
      
      // If we copied the data, or our last time stamp is wrong, read it in
      if (newdata || (havelocal && lastUpdate_ != local.m_mtime))
      {
	lastUpdate_ = local.m_mtime;
	reset();
	try
	{
	  if (db_)
	    dbclose(&db_);

	  int rc;
	  if ((rc = sqlite3_open(localFileName_, &db_)) != SQLITE_OK)
	    throw dberror("sqlite3_open_v2", rc, true, &db_,
			  StringFormat("failed to open database %1")
			  .arg(localFileName_.name()));

	  fillMap("datasets", idToDataset_);
	  fillMap("steps", idToStep_);
	  fillMap("systems", idToSystem_);
	  fillMap("categories", idToCategory_);
	  fillMap("files", idToFile_);
	  fillMap("names", idToName_);

	  IdKeyMap::iterator i, e;
	  for (i = idToDataset_.begin(), e = idToDataset_.end(); i != e; ++i)
	    datasetToId_[i->second] = i->first;

	  for (i = idToStep_.begin(), e = idToStep_.end(); i != e; ++i)
	    stepToId_[i->second] = i->first;

	  char *errmsg = 0;
	  if ((rc = sqlite3_exec(db_, "select distinct runnr, dataset, step from t_data",
				 &setupRunData, this, &errmsg)) != SQLITE_OK)
	    throw dberror("sqlite3_exec", rc, errmsg);
	}
	catch (Error &e)
	{
	  logerr()
	    << "failed to read database " << localFileName_
	    << ": " << e.explain() << std::endl;
	  reset();
	}
	catch (std::exception &e)
	{
	  logerr()
	    << "failed to read database " << localFileName_
	    << ": " << e.what() << std::endl;
	  reset();
	}
      }

      // Check if the run is known in the database.  Bail out if not.
      RunCache::iterator pos = runCache_.find(runkey);
      if (! db_ || pos == runCache_.end())
	return;

      // It's known.  Read the data in if we haven't done so yet.
      if (pos->second.stamp == 0)
      {
	try
	{
	  loginfo() << "reading data for run key " << runkey << std::endl;
	  RunData &d = pos->second;
	  StringList k = StringOps::split(runkey, ':');
	  ASSERT(d.runId == atol(k[0].c_str()));
	  ASSERT(d.datasetId == datasetToId_[k[1]]);
	  ASSERT(d.stepId == stepToId_[k[2]]);
	  d.stamp = Time::current();
	  d.runNumber = 0;
	  d.lumiNumber = 0;
	  d.eventNumber = 0;

	  int rc;
	  sqlite3_stmt *stmt = 0;
	  if ((rc = sqlite3_prepare_v2(db_,
				       "select * from t_data\n"
				       "where runnr = ?\n"
				       " and dataset = ?\n"
				       " and step = ?",
				       -1, &stmt, 0))
	      != SQLITE_OK)
	    throw dberror("sqlite3_prepare_v2", rc, false, &db_,
			  "failed to read run data from the database");

	  if ((rc = sqlite3_bind_text(stmt, 1, k[0].c_str(), -1, SQLITE_STATIC)) != SQLITE_OK
	      || (rc = sqlite3_bind_int(stmt, 2, d.datasetId)) != SQLITE_OK
	      || (rc = sqlite3_bind_int(stmt, 3, d.stepId)) != SQLITE_OK)
	    throw dberror("sqlite3_bind_text", rc, false, &db_,
			  "failed to bind run key arguments");

	  while ((rc = sqlite3_step(stmt)) != SQLITE_DONE)
	  {
	    if (rc != SQLITE_ROW)
	      throw dberror("sqlite3_step", rc, db_, stmt);

	    std::string name;
	    const unsigned char *datastr = sqlite3_column_text(stmt, 11);
	    const std::string &sysname = idToSystem_[sqlite3_column_int(stmt, 4)];
	    const std::string &objname = idToName_[sqlite3_column_int(stmt, 6)];
	    name.reserve(sysname.size() + objname.size() + 2);
	    name += sysname;
	    name += '/';
	    name += objname;

	    if (datastr && name.find("/EventInfo/i") != std::string::npos)
	      updateRunInfo((const char *) datastr,
			    d.runNumber,
			    d.lumiNumber,
			    d.eventNumber);

	    shared_ptr<VisDQMItem> i(new VisDQMItem);
	    i->fileid = sqlite3_column_int(stmt, 7);
	    i->flags = sqlite3_column_int(stmt, 8);
	    i->version = lastUpdate_.ns();
	    i->name = StringAtom(&stree, name);
	    i->plotter = this;
	    switch (*sqlite3_column_text(stmt, 10))
	    {
	    case 'c':
	      i->flags |= DQM_FLAG_HAS_REFERENCE;
	      /* fall through */

	    case 'C':
	      i->realname = StringFormat("Run %1/%2/%3/%4")
			    .arg(sqlite3_column_int(stmt, 1))
			    .arg(idToSystem_[sqlite3_column_int(stmt, 4)])
			    .arg(idToCategory_[sqlite3_column_int(stmt, 5)])
			    .arg(idToName_[sqlite3_column_int(stmt, 6)]);
	      break;

	    case 's':
	      i->flags |= DQM_FLAG_HAS_REFERENCE;
	      /* fall through */

	    case 'S':
	      i->realname = name;
	      break;

	    default:
	      logwarn() << "unrecognised name scheme '"
			<< *sqlite3_column_text(stmt, 10)
			<< "'\n";
	      i->realname.clear();
	      break;
	    }
	    if (datastr)
	      i->data.insert(i->data.end(), datastr,
			     datastr + strlen((const char *) datastr));
	    buildParentNames(i->parents, i->name);
	    d.items[i->name] = i;
	  }

	  sqlite3_finalize(stmt);
	}
	catch (Error &e)
	{
	  logerr()
	    << "failed to read database " << localFileName_
	    << ": " << e.explain() << std::endl;
	  reset();
	}
	catch (std::exception &e)
	{
	  logerr()
	    << "failed to read database " << localFileName_
	    << ": " << e.what() << std::endl;
	  reset();
	}
      }
    }

  static void
  fillRunKey(std::string &runkey,
	     const std::string &dataset,
	     const std::string &step,
	     const std::string &runnr)
    {
      runkey.reserve(runnr.size() + dataset.size() + step.size() + 4);
      runkey += runnr; runkey += ':';
      runkey += dataset; runkey += ':';
      runkey += step;
    }

  virtual void
  prepareSession(py::dict session)
    {
      if (! hasSessionKey(session, "dqm.dataset")
	  && ! hasSessionKey(session, "dqm.step")
	  && ! hasSessionKey(session, "dqm.runnr"))
      {
	StringList runkeys;

	{
	  Lock gate(&lock_);
	  update(std::string());
	  runkeys.reserve(runCache_.size());
	  for (RunCache::iterator i = runCache_.begin(), e = runCache_.end(); i != e; ++i)
	    runkeys.push_back(i->first);
	}

	std::sort(runkeys.begin(), runkeys.end(), natcmpstr);
	if (! runkeys.empty())
	{
	  StringList k = StringOps::split(runkeys.back(), ':');
	  session["dqm.runnr"] = k[0];
	  session["dqm.dataset"] = k[1];
	  session["dqm.step"] = k[2];
	}
      }
    }

  // Extract data from this source.  Calls update() then simply copies
  // into the result the objects that don't already exist there.
  virtual void
  scan(VisDQMItems &result,
       StringSet &datasets,
       StringSet &steps,
       StringSet &runs,
       const std::string &dataset,
       const std::string &step,
       const std::string &runnr,
       long &curRunNr,
       long &curLumiNr,
       long &curEventNr,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm)
    {
      std::string runkey;
      fillRunKey(runkey, dataset, step, runnr);

      Lock gate(&lock_);
      update(runkey);
      KeyIdMap::iterator ki, ke;
      RunCache::iterator ri, re;
      for (ki = datasetToId_.begin(), ke = datasetToId_.end(); ki != ke; ++ki)
	datasets.insert(ki->first);
      for (ki = stepToId_.begin(), ke = stepToId_.end(); ki != ke; ++ki)
	steps.insert(ki->first);
      for (ri = runCache_.begin(), re = runCache_.end(); ri != re; ++ri)
      {
	char buf[32];
	sprintf(buf, "%ld", ri->second.runId);
	runs.insert(buf);

	if (ri->first != runkey)
	  continue;

	RunData &d = ri->second;
	curRunNr = d.runNumber;
	curLumiNr = d.lumiNumber;
	curEventNr = d.eventNumber;

	VisDQMItems::iterator oi, oe;
        result.resize(result.size() + d.items.size());
	for (oi = d.items.begin(), oe = d.items.end(); oi != oe; ++oi)
	{
	  VisDQMItem &o = *oi->second;

	  if (rxmatch && ! rxmatch->match(o.name.string()))
	    continue;

	  if (rxsearch && rxsearch->search(o.name.string()) < 0)
	    continue;

	  if (alarm && ((o.flags & HAS_ALARM) != 0) != (*alarm == true))
	    continue;

	  result[o.name] = oi->second;
	}
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMWorkspace
{
protected:
  py::object	gui_;
  double	guiTimeStamp_;
  std::string	guiServiceName_;
  std::string	guiServiceListJSON_;
  std::string	guiWorkspaceListJSON_;
  std::string	workspaceName_;

public:
  VisDQMWorkspace(py::object gui, const std::string &name)
    : gui_(gui),
      guiTimeStamp_(py::extract<double>(gui.attr("stamp"))),
      guiServiceName_(py::extract<std::string>(gui.attr("serviceName"))),
      workspaceName_(name)
    {
    }

  virtual ~VisDQMWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      return "{}";
    }

protected:
  template <class T> T
  workspaceParam(const py::dict &session, const char *key)
    {
      py::dict d = py::extract<py::dict>(session.get(key));
      return py::extract<T>(d.get(this->workspaceName_));
    }
      
  template <class T> T
  workspaceParam(const py::dict &session, const char *key, const T &def)
    {
      py::dict d = py::extract<py::dict>(session.get(key));
      return py::extract<T>(d.get(this->workspaceName_, def));
    }

  // Get the list of native VisDQMSources configured in the GUI.
  void
  sources(std::vector<VisDQMSource *> &into)
    {
      py::list sources = py::extract<py::list>(gui_.attr("sources"));
      py::stl_input_iterator<py::object> i(sources), e;
      for ( ; i != e; ++i)
      {
	py::extract<VisDQMSource *> x(*i);
	if (x.check())
	  into.push_back(x());
      }
    }

  // Build and return a cached JSON representation of the DQM GUI
  // service list.  This must be called when it's safe to access
  // python.  Will build the list just once.
  const std::string &
  serviceListJSON(void)
    {
      if (guiServiceListJSON_.empty())
      {
	StringList parts;
	parts.reserve(100);

	py::list services = py::extract<py::list>(gui_.attr("services"));
	for (py::stl_input_iterator<py::tuple> i(services), e; i != e; ++i)
	  parts.push_back(StringFormat("{title: '%1', href: '%2'}")
			  .arg(py::extract<std::string>((*i)[0]))
			  .arg(py::extract<std::string>((*i)[1])));
	guiServiceListJSON_ = StringOps::join(parts, ", ");
      }

      return guiServiceListJSON_;
    }

  // Build and return a cached JSON representation of the DQM GUI
  // workspace object list.  This must be called when it's safe to
  // access python.  Will build the list just once.
  const std::string &
  workspaceListJSON(void)
    {
      if (guiWorkspaceListJSON_.empty())
      {
	StringList parts;
	parts.reserve(100);
	py::list workspaces = py::extract<py::list>(gui_.attr("workspaces"));
	for (py::stl_input_iterator<py::object> i(workspaces), e; i != e; ++i)
	{
	  int         rank(py::extract<int>(i->attr("rank")));
	  std::string name(py::extract<std::string>(i->attr("name")));
	  std::string category(py::extract<std::string>(i->attr("category")));
	  std::string lower(name);
	  std::transform(lower.begin(), lower.end(), lower.begin(), tolower);
	  parts.push_back(StringFormat("{title: '%1', label: '%2', category: '%3', rank: %4}")
			  .arg(name).arg(lower).arg(category).arg(rank));
	}
	guiWorkspaceListJSON_ = StringOps::join(parts, ", ");
      }

      return guiWorkspaceListJSON_;
    }

  // Get the contents currently shown in this session.
  void
  buildContents(const std::vector<VisDQMSource *> &sources,
		VisDQMItems &contents,

		StringAtomSet &myobjs,
		StringSet &datasets,
		StringSet &steps,
		StringSet &runs,
		
		const std::string &dataset,
		const std::string &step,
		const std::string &runnr,

		long &curRunNr,
		long &curLumiNr,
		long &curEventNr,

		Regexp *rxmatch,
		Regexp *rxsearch,
		bool *alarm)
    {
      // Now get filtered contents from all the sources.
      for (size_t i = 0, e = sources.size(); i != e; ++i)
	sources[i]->scan(contents, datasets, steps, runs,
			 dataset, step, runnr,
			 curRunNr, curLumiNr, curEventNr,
			 rxmatch, rxsearch, alarm);
      

      // Add in the quick collection, filtered against rx if there is
      // one, but not the general workspace filter.
      StringAtomSet::iterator mi, me;
      for (mi = myobjs.begin(), me = myobjs.end(); mi != me; ++mi)
      {
	std::string trailer(mi->string(), StringOps::rfind(mi->string(), '/')+1);
	StringAtom label(&stree, "Quick collection/" + trailer);
	if (! rxsearch || rxsearch->search(label.string()) >= 0)
	{
	  shared_ptr<VisDQMItem> i(new VisDQMItem);
	  i->fileid = 0;
	  i->flags = 0;
	  i->version = 0;
	  i->name = label;
	  i->plotter = 0;
	  buildParentNames(i->parents, i->name);

	  VisDQMItems::iterator pos = contents.find(*mi);
	  if (pos != contents.end() && pos->second->layout)
	    i->layout = pos->second->layout;
	  else
	  {
	    shared_ptr<VisDQMLayoutRows> layout(new VisDQMLayoutRows);
	    shared_ptr<VisDQMLayoutRow> row(new VisDQMLayoutRow);
	    shared_ptr<VisDQMLayoutItem> column(new VisDQMLayoutItem);
	    column->path = *mi;
	    layout->push_back(row);
	    row->columns.push_back(column);
	    i->layout = layout;
	  }

	  contents[label] = i;
	}
      }
    }

  // Utility method to build a look-up table of shown objects.
  void
  buildShown(const VisDQMItems &contents,
	     StringAtomSet &shown,
	     const std::string &root,
	     const std::string &picked,
	     const std::string &focusstr)
    {
      std::string myroot;
      if (! root.empty())
      {
	myroot.reserve(root.size() + 2);
	myroot += root;
	myroot += '/';
      }

      if (! picked.empty())
      {
	size_t slash;
	if (contents.count(StringAtom(&stree, picked, StringAtom::TestOnly)) == 0)
	{
	  myroot.reserve(picked.size() + 2);
	  if ((slash = picked.find_last_not_of('/')) == std::string::npos)
	    myroot = picked;
	  else
	    myroot = std::string(picked, 0, slash+1);
	  myroot += '/';
	}
	else if ((slash = picked.rfind('/')) != std::string::npos)
	  myroot = std::string(picked, 0, slash+1);
	else
	  myroot = "";
      }
      
      StringAtom focus(&stree, focusstr);
      if (! focusstr.empty() && contents.count(focus))
	shown.insert(focus);
      else
      {
	VisDQMItems::const_iterator ci, ce;
	for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	  if (ci->first.string().size() > myroot.size()
	      && ci->first.string().compare(0, myroot.size(), myroot) == 0
	      && ci->first.string().find('/', myroot.size()) == std::string::npos)
	    shown.insert(ci->first);
      }
    }

  // Update the object metrics for the items in @a contents.
  // "Shown" means the object is included in canvas listing.
  void
  updateStatus(const VisDQMItems &contents,
	       VisDQMStatusMap &status)
    {
      VisDQMStatus empty;
      empty.isdir = 0;
      empty.nleaves = 0;
      empty.nalarm = 0;
      empty.nlive = 0;

      VisDQMStatus leaf;
      leaf.isdir = 0;
      leaf.nleaves = 1;
      leaf.nalarm = 0;
      leaf.nlive = 0;

      status[StringAtom(&stree, "")] = empty;
      status[StringAtom(&stree, "Quick collection")] = empty;

      status.resize(10*contents.size());
      VisDQMItems::const_iterator ci, ce, lpos;
      for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
      {
	// Initialise object data.
	VisDQMItem &o = *ci->second;
	VisDQMStatus &s = status[ci->first];
	s.nleaves = 1;
	s.nalarm = ((o.flags & HAS_ALARM) ? 1 : 0);
	s.nlive = ((o.flags & DQMNet::DQM_FLAG_DEAD) ? 0 : 1);

	// If this is a layout, resolve it.
	if (o.layout)
	{
	  VisDQMLayoutRows &rows = *o.layout;
	  for (size_t nrow = 0; nrow < rows.size(); ++nrow)
	  {
	    VisDQMLayoutRow &row = *rows[nrow];
	    for (size_t ncol = 0; ncol < row.columns.size(); ++ncol)
	    {
	      VisDQMLayoutItem &col = *row.columns[ncol];
	      if ((lpos = contents.find(col.path)) != contents.end())
		s.nalarm += ((lpos->second->flags & HAS_ALARM) ? 1 : 0);
	      else if (! col.path.string().empty())
		s.nlive = 0;
	    }
	  }
	}

	// Update data to "virtual parents."
	for (size_t i = 0, e = o.parents.size(); i != e; ++i)
	{
	  VisDQMStatusMap::iterator pos = status.find(o.parents[i]);
	  if (pos == status.end())
	    pos = status.insert(VisDQMStatusMap::value_type(o.parents[i], empty)).first;
	  pos->second.nleaves += s.nleaves;
	  pos->second.nalarm += s.nalarm;
	  pos->second.nlive += s.nlive;
	  pos->second.isdir = 1;
	}
      }
    }

  static std::string
  runListToJSON(const StringSet &runs)
    {
      std::string result;
      result.reserve(20 * runs.size());

      StringSet::const_reverse_iterator i, e;
      for (i = runs.rbegin(), e = runs.rend(); i != e; ++i)
      {
	if (! result.empty())
	  result += ", ";

	result += '"';
	result += *i;
	result += '"';
      }

      return result;
    }

private:
  VisDQMWorkspace(VisDQMWorkspace &);
  VisDQMWorkspace &operator=(VisDQMWorkspace &);
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMSummaryWorkspace : public VisDQMWorkspace
{
  typedef std::map<std::string, std::string> KeyValueMap;
  typedef std::map<std::string, KeyValueMap> SummaryMap;

  Regexp rxevi_;
  Regexp rxvar_;
public:
  VisDQMSummaryWorkspace(py::object gui, const std::string &name)
    : VisDQMWorkspace(gui, name),
      rxevi_("(.*?)/EventInfo/(.*)"),
      rxvar_("<.*?>(i|f|s|qr)=(.*?)</.*?>")
    {
      rxevi_.study();
      rxvar_.study();
    }

  virtual ~VisDQMSummaryWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      std::string dataset(py::extract<std::string>(session.get("dqm.dataset")));
      std::string step(py::extract<std::string>(session.get("dqm.step")));
      std::string runnr(py::extract<std::string>(session.get("dqm.runnr")));
      std::string qplot(py::extract<std::string>(session.get("dqm.qplot", "")));

      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringAtomSet		myobjs; // deliberately empty, not needed
	StringSet		datasets;
	StringSet		steps;
	StringSet		runs;
	VisDQMItems::iterator	ci, ce;
	VisDQMStatusMap		status;
	RegexpMatch		m;
	std::string		plotter;
	SummaryMap		summary;
	long			curRunNr = 0;
	long			curLumiNr = 0;
	long			curEventNr = 0;

	buildContents(srclist, contents, myobjs,
		      datasets, steps, runs,
		      dataset, step, runnr,
		      curRunNr, curLumiNr, curEventNr,
		      0, 0, 0);

	updateStatus(contents, status);

	for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	{
	  VisDQMItem &o = *ci->second;
	  if (! (o.flags & DQMNet::DQM_FLAG_SCALAR))
	    continue;

	  m.reset();
	  if (! rxevi_.match(o.name.string(), 0, 0, &m))
	    continue;

	  std::string system(m.matchString(o.name.string(), 1));
	  std::string param(m.matchString(o.name.string(), 2));
	  KeyValueMap &info = summary[system];
	  if (info.empty())
	  {
	    plotter = StringFormat("%1/%2%3/%4")
		      .arg(o.plotter ? o.plotter->plotter() : "unknown")
		      .arg(runnr)
		      .arg(dataset)
		      .arg(step);
	    info["MEs"] = StringFormat("%1").arg(status[StringAtom(&stree, system)].nleaves);
	    info["processName"] = system;
	  }

	  m.reset();
	  if (rxvar_.match((const char *) &o.data[0], 0, 0, &m))
	    info[param] = m.matchString((const char *) &o.data[0], 2);
	}

	return StringFormat("([{kind: 'AutoUpdate', interval: %1, stamp: %2, serverTime: %15},"
			    "{kind: 'DQMHeaderRow', run: \"%3\", lumi: \"%4\", event: \"%5\","
			    " service: '%6', workspace: '%7', page: 1, pages: 1,"
			    " services: [%8], workspaces: [%9], runs: [%10], runid: %11},"
			    "{kind: 'DQMSummary', items: [%12], qrender: '%13', qplot: '%14'}])")
	  .arg(runnr == "Live" ? 30 : 300)
	  .arg(guiTimeStamp_, 0, 'f')
	  .arg(curRunNr == 0 ? std::string("(Waiting)")
	       : thousands(StringFormat("%1").arg(curRunNr)))
	  .arg(curLumiNr == 0 ? std::string("_")
	       : thousands(StringFormat("%1").arg(curLumiNr)))
	  .arg(curEventNr == 0 ? std::string("_")
	       : thousands(StringFormat("%1").arg(curEventNr)))
	  .arg(guiServiceName_)
	  .arg(workspaceName_)
	  .arg(services)
	  .arg(workspaces)
	  .arg(runListToJSON(runs))
	  .arg(curRunNr == 0 ? std::string("'Live'") : StringFormat("%1").arg(curRunNr))
	  .arg(summaryToJSON(summary))
	  .arg(plotter)
	  .arg(qplot)
	  .arg((Time::current() - startTime).ns() * 1e-6, 0, 'f');
      }
    }

private:
  static std::string
  summaryToJSON(const SummaryMap &summary)
    {
      size_t len = 4 * summary.size();
      SummaryMap::const_iterator si, se;
      KeyValueMap::const_iterator ki, ke;
      for (si = summary.begin(), se = summary.end(); si != se; ++si)
      {
	len += 8 * si->second.size();
	for (ki = si->second.begin(), ke = si->second.end(); ki != ke; ++ki)
	  len += ki->first.size() + 2 * ki->second.size();
      }

      std::string result;
      result.reserve(len + 1);
      for (si = summary.begin(), se = summary.end(); si != se; ++si)
      {
	if (! result.empty())
	  result += ", ";
	result += "{";

	bool first = true;
	for (ki = si->second.begin(), ke = si->second.end(); ki != ke; ++ki, first = false)
	{
	  if (! first)
	    result += ", ";
	  result += stringToJSON(ki->first);
	  result += ": ";
	  result += stringToJSON(ki->second);
	}
	
	result += "}";
      }

      return result;
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMQualityWorkspace : public VisDQMWorkspace
{
  struct QMapItem
  {
    std::string location;
    std::string name;
    std::string	reportSummary;
    std::string eventTimeStamp;
    uint64_t version;
  };

  typedef std::map<std::string, QMapItem> QMap;
  Regexp rxvar_;
public:
  VisDQMQualityWorkspace(py::object gui, const std::string &name)
    : VisDQMWorkspace(gui, name),
      rxvar_("<.*?>(i|f|s|qr)=(.*?)</.*?>")
    {}

  virtual ~VisDQMQualityWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      std::string dataset(py::extract<std::string>(session.get("dqm.dataset")));
      std::string step(py::extract<std::string>(session.get("dqm.step")));
      std::string runnr(py::extract<std::string>(session.get("dqm.runnr")));

      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringAtomSet		myobjs; // deliberately empty, not needed
	StringSet		datasets;
	StringSet		steps;
	StringSet		runs;
	VisDQMItems::iterator	ci, ce;
	std::string		plotter;
	std::string		label;
	QMap			qmap;
	long			curRunNr = 0;
	long			curLumiNr = 0;
	long			curEventNr = 0;
	RegexpMatch		m;

	buildContents(srclist, contents, myobjs,
		      datasets, steps, runs,
		      dataset, step, runnr,
		      curRunNr, curLumiNr, curEventNr,
		      0, 0, 0);

	for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	{
	  VisDQMItem &o = *ci->second;
	  size_t pos = o.name.string().find("/EventInfo/reportSummaryMap");
	  if (pos > 0 && pos != std::string::npos)
	  {
	    size_t start = o.name.string().rfind('/', pos ? pos-1 : 0);
	    if (start == std::string::npos) start = 0;
	    label = o.name.string().substr(start, pos-start);
	    QMapItem &i = qmap[label];
	    i.location = StringFormat("%1/%2%3/%4")
			 .arg(o.plotter ? o.plotter->plotter() : "unknown")
			 .arg(runnr)
			 .arg(dataset)
			 .arg(step);
	    i.name = o.name.string();
	    i.version = o.version;
	    i.reportSummary.clear();
	    i.eventTimeStamp.clear();

	    VisDQMItems::iterator other;
	    StringAtom reportSummary(&stree,
				     label + "/EventInfo/reportSummary",
				     StringAtom::TestOnly);
	    StringAtom eventTimeStamp(&stree,
				      label + "/EventInfo/eventTimeStamp",
				      StringAtom::TestOnly);

	    m.reset();
	    other = contents.find(reportSummary);
	    if (other != ce
		&& (other->second->flags & DQMNet::DQM_FLAG_SCALAR)
		&& other->second->data.size() > 0
		&& rxvar_.match((const char *) &other->second->data[0], 0, 0, &m))
	      i.reportSummary.append((const char *) &other->second->data[0]
				     + m.matchPos(2), m.matchLength(2));

	    m.reset();
	    other = contents.find(eventTimeStamp);
	    if (other != ce
		&& (other->second->flags & DQMNet::DQM_FLAG_SCALAR)
		&& other->second->data.size() > 0
		&& rxvar_.match((const char *) &other->second->data[0], 0, 0, &m))
	      i.eventTimeStamp.append((const char *) &other->second->data[0]
				      + m.matchPos(2), m.matchLength(2));
	  }
	}

	return StringFormat("([{kind: 'AutoUpdate', interval: %1, stamp: %2, serverTime: %13},"
			    "{kind: 'DQMHeaderRow', run: \"%3\", lumi: \"%4\", event: \"%5\","
			    " service: '%6', workspace: '%7', page: 1, pages: 1,"
			    " services: [%8], workspaces: [%9], runs: [%10], runid: %11},"
			    "{kind: 'DQMQuality', items: [%12]}])")
	  .arg(runnr == "Live" ? 30 : 300)
	  .arg(guiTimeStamp_, 0, 'f')
	  .arg(curRunNr == 0 ? std::string("(Waiting)")
	       : thousands(StringFormat("%1").arg(curRunNr)))
	  .arg(curLumiNr == 0 ? std::string("_")
	       : thousands(StringFormat("%1").arg(curLumiNr)))
	  .arg(curEventNr == 0 ? std::string("_")
	       : thousands(StringFormat("%1").arg(curEventNr)))
	  .arg(guiServiceName_)
	  .arg(workspaceName_)
	  .arg(services)
	  .arg(workspaces)
	  .arg(runListToJSON(runs))
	  .arg(curRunNr == 0 ? std::string("'Live'") : StringFormat("%1").arg(curRunNr))
	  .arg(qmapToJSON(qmap))
	  .arg((Time::current() - startTime).ns() * 1e-6, 0, 'f');
      }
    }

private:
  static std::string
  qmapToJSON(const QMap &qmap)
    {
      size_t len = 30 * qmap.size();
      QMap::const_iterator i, e;
      for (i = qmap.begin(), e = qmap.end(); i != e; ++i)
	len += 80 + 2 * (i->first.size()
			 + i->second.location.size()
			 + i->second.name.size()
			 + i->second.reportSummary.size()
			 + i->second.eventTimeStamp.size());

      std::string result;
      result.reserve(len);
      for (i = qmap.begin(), e = qmap.end(); i != e; ++i)
      {
	char buf[32];
	__extension__
	  sprintf(buf, "%ju", (uintmax_t) i->second.version);
	if (! result.empty())
	  result += ", ";
	result += "{label:";
	result += stringToJSON(i->first);
	result += ",version:";
	result += buf;
	result += ",name:";
	result += stringToJSON(i->second.name);
	result += ",location:";
	result += stringToJSON(i->second.location);
	result += ",reportSummary:";
	result += stringToJSON(i->second.reportSummary);
	result += ",eventTimeStamp:";
	result += stringToJSON(i->second.eventTimeStamp);
	result += "}";
      }

      return result;
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMContentWorkspace : public VisDQMWorkspace
{
  shared_ptr<Regexp> rxmatch_;
public:
  VisDQMContentWorkspace(py::object gui,
			 const std::string &name,
			 const std::string &match)
    : VisDQMWorkspace(gui, name),
      rxmatch_(rx(match))
    {}

  virtual ~VisDQMContentWorkspace(void)
    {}

  static void
  translate(const py::dict &d, VisDQMDrawOptionMap &drawopts)
    {
    }

  // Helper function to present session state.  Returns a JSON object
  // representing current menu and canvas state.
  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      std::string dataset      (py::extract<std::string>(session.get("dqm.dataset")));
      std::string step         (py::extract<std::string>(session.get("dqm.step")));
      std::string runnr        (py::extract<std::string>(session.get("dqm.runnr")));
      std::string root         (workspaceParam<std::string>(session, "dqm.root", ""));
      std::string picked       (workspaceParam<std::string>(session, "dqm.picked", "")); // None
      std::string focus        (workspaceParam<std::string>(session, "dqm.focus", ""));  // None
      std::string filter       (py::extract<std::string>(session.get("dqm.filter")));
      std::string size	       (py::extract<std::string>(session.get("dqm.size")));
      std::string rxstr        (py::extract<std::string>(session.get("dqm.search")));
      std::string rxerr;
      int         perpage      (py::extract<int>(session.get("dqm.perpage")));
      int         page         (workspaceParam<int>(session, "dqm.page", 1));
      int         playmode     (py::extract<int>(session.get("dqm.playmode")));
      int         playpos      (py::extract<int>(session.get("dqm.playpos")));
      int         playinterval (py::extract<int>(session.get("dqm.playinterval")));
      py::dict    pydrawopts   (workspaceParam<py::dict>(session, "dqm.drawopts", py::dict()));
      StringAtomSet myobjs;

      for (py::stl_input_iterator<std::string> e, i (workspaceParam<py::tuple>(session, "dqm.myobjs", py::tuple())); i != e; ++i)
        myobjs.insert(StringAtom(&stree, *i));

      VisDQMDrawOptionMap drawopts;
      for (py::stl_input_iterator<std::string> i(pydrawopts), e; i != e; ++i)
	translateDrawOptions(py::extract<py::dict>(pydrawopts.get(*i)),
			     drawopts[StringAtom(&stree, *i)]);

      shared_ptr<Regexp> rxsearch;
      try
      {
	if (! rxstr.empty())
	  rxsearch = rx(rxstr, Regexp::IgnoreCase);
      }
      catch (Error &e)
      {
	rxerr = e.explainSelf();
      }
      catch (std::exception &e)
      {
	rxerr = e.what();
      }
      
      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringSet		datasets;
	StringSet		steps;
	StringSet		runs;
	StringAtomSet		shown;
	VisDQMItems::iterator	ci, ce;
	VisDQMStatusMap		status;
	RegexpMatch		m;
	std::string		data;
	long			curRunNr = 0;
	long			curLumiNr = 0;
	long			curEventNr = 0;
	bool			alarms = false;
	bool			*qalarm = 0;

	if (filter == "all")
	  qalarm = 0;
	else if (filter == "alarms")
	  alarms = true, qalarm = &alarms;
	else if (filter == "nonalarms")
	  alarms = false, qalarm = &alarms;

	buildContents(srclist, contents, myobjs,
		      datasets, steps, runs,
		      dataset, step, runnr,
		      curRunNr, curLumiNr, curEventNr,
		      rxmatch_.get(), rxsearch.get(), qalarm);
	buildShown(contents, shown, root, picked, focus);
	updateStatus(contents, status);

	if (! status.count(StringAtom(&stree, root)))
	  root = "";
	if (! picked.empty() && ! status.count(StringAtom(&stree, picked)))
	  picked = ""; // None
	if (! focus.empty() && ! status.count(StringAtom(&stree, focus)))
	  focus = ""; // None

	int pages = (int(shown.size()) + perpage - 1) / perpage;
	page = std::min(std::max(1, page), pages);
	if (playmode && shown.size())
	{
	  playpos = std::min(playpos, int(shown.size())-1);
	  page = std::min(std::max(1, playpos/perpage+1), pages);
	}
	else
	  playpos = 0;

	return StringFormat("([{kind: 'AutoUpdate', interval: %1, stamp: %2, serverTime: %30},"
			    "{kind: 'DQMHeaderRow', run: \"%3\", lumi: \"%4\", event: \"%5\","
			    " service: '%6', workspace: '%7', page: %8, pages: %9,"
			    " services: [%10], workspaces: [%11], runs: [%12], runid: %13},"
			    "{kind: 'DQMList', quick: %14, tree: %15,"
			    " def: {dataset: [%16], step: [%17], filter: '%18', search: %19}},"
			    "{kind: 'DQMCanvas', items: %20, params: {perpage: %21, picked: %22,"
			    " focus: %23, size: '%24', play: { on: %25, interval: %26, pos: %27,"
			    " page: %28, max: %29 }}}])")
	  .arg((playmode || runnr != "Live") ? 300 : 30)
	  .arg(guiTimeStamp_, 0, 'f')
	  .arg(curRunNr == 0 ? std::string("(Waiting)")
	       : thousands(StringFormat("%1").arg(curRunNr)))
	  .arg(curLumiNr == 0 ? std::string("_")
	       : thousands(StringFormat("%1").arg(curLumiNr)))
	  .arg(curEventNr == 0 ? std::string("_")
	       : thousands(StringFormat("%1").arg(curEventNr)))
	  .arg(guiServiceName_)
	  .arg(workspaceName_)
	  .arg(page)
	  .arg(pages)
	  .arg(services)
	  .arg(workspaces)
	  .arg(runListToJSON(runs))
	  .arg(curRunNr == 0 ? std::string("'Live'") : StringFormat("%1").arg(curRunNr))
	  .arg(treeToJSON(contents, status, StringAtom(&stree, "Quick collection"), "/", focus, picked, 2))
	  .arg(treeToJSON(contents, status, StringAtom(&stree, root), "Quick collection", focus, picked, 2))
	  .arg(stringSetToJSON(datasets, dataset))
	  .arg(stringSetToJSON(steps, step))
	  .arg(filter)
	  .arg(StringFormat("{pattern: %1, valid: %2, nmatches: %3, error: %4}")
	       .arg(stringToJSON(rxstr))
	       .arg(rxerr.empty() ? 1 : 0)
	       .arg(contents.size())
	       .arg(stringToJSON(rxerr)))
	  .arg(shownToJSON(contents, status, drawopts,
			   dataset, step, runnr,
			   shown, perpage*std::max(0, page-1), perpage*page))
	  .arg(perpage)
	  .arg(stringToJSON(picked, true))
	  .arg(! focus.empty() ? 1 : 0)
	  .arg(! focus.empty() && shown.size() ? std::string("XL") : size)
	  .arg(playmode)
	  .arg(playinterval)
	  .arg(playpos)
	  .arg(std::max(0, page-1))
	  .arg(shown.size())
	  .arg((Time::current() - startTime).ns() * 1e-6, 0, 'f');
      }
    }

  std::string
  stringSetToJSON(const StringSet &s, const std::string &selected)
    {
      StringList sorted(s.begin(), s.end());
      std::sort(sorted.begin(), sorted.end(), natcmpstr);

      size_t size = 0;
      StringList::iterator i, e;
      for (i = sorted.begin(), e = sorted.end(); i != e; ++i)
	size += 25 + i->size();

      std::string result;
      result.reserve(size);
      for (i = sorted.begin(), e = sorted.end(); i != e; ++i)
      {
	if (! result.empty())
	  result += ", ";

	result += "{name: '";
	result += *i;
	result += "', selected: ";
	result += (*i == selected ? '1' : '0');
	result += '}';
      }

      return result;
    }

  static std::string
  axisToJSON(const VisDQMAxisOptions &axis)
    {
      return StringFormat("{min: %1, max: %2, type: %3}")
	.arg(stringToJSON(axis.min, true))
	.arg(stringToJSON(axis.max, true))
	.arg(stringToJSON(axis.type));
    }

  static std::string
  layoutToJSON(const VisDQMShownItem &x,
	       const std::string &dataset,
	       const std::string &step,
	       const std::string &runnr)
    {
      char vbuf[64]; __extension__ sprintf(vbuf, "%ju", (uintmax_t) x.version);

      return StringFormat("{name: %1, desc: %2, location: \"%3/%4%5/%6\","
			  " version: \"%7\", alarm: %8, live: %9,"
			  " xaxis: %10, yaxis: %11, zaxis: %12,"
			  " drawopts: %13, withref: %14}")
	.arg(stringToJSON(x.name.string(), true))
	.arg(stringToJSON(x.desc, true))
	.arg(x.plotter ? x.plotter->plotter() : "unknown")
	.arg(runnr)
	.arg(dataset)
	.arg(step)
	.arg(vbuf)
	.arg(x.nalarm)
	.arg(x.nlive)
	.arg(axisToJSON(x.drawopts.xaxis))
	.arg(axisToJSON(x.drawopts.yaxis))
	.arg(axisToJSON(x.drawopts.zaxis))
	.arg(stringToJSON(x.drawopts.drawopts))
	.arg(stringToJSON(x.drawopts.withref));
    }

  static std::string
  layoutToJSON(const VisDQMShownRow &layout,
	       const std::string &dataset,
	       const std::string &step,
	       const std::string &runnr)
    {
      VisDQMShownItems::const_iterator i, e;
      std::string result;
      result.reserve(2048);
      result += '[';
      for (size_t i = 0, e = layout.columns.size(); i != e; ++i)
      {
	if (i > 0)
	  result += ", ";
	result += layoutToJSON(*layout.columns[i], dataset, step, runnr);
      }
      result += ']';
      return result;
    }

  static std::string
  layoutToJSON(const VisDQMShownRows &layout,
	       const std::string &dataset,
	       const std::string &step,
	       const std::string &runnr)
    {
      VisDQMShownRows::const_iterator i, e;
      std::string result;
      result.reserve(2048);
      result += '[';
      for (size_t i = 0, e = layout.size(); i != e; ++i)
      {
	if (i > 0)
	  result += ", ";
	result += layoutToJSON(*layout[i], dataset, step, runnr);
      }
      result += ']';
      return result;
    }

  static std::string
  shownToJSON(const VisDQMItems &contents,
	      const VisDQMStatusMap &status,
	      const VisDQMDrawOptionMap &drawopts,
	      const std::string &dataset,
	      const std::string &step,
	      const std::string &runnr,
	      const StringAtomSet &shown,
	      int first,
	      int last)
    {
      StringAtomList shownidx(shown.begin(), shown.end());
      std::sort(shownidx.begin(), shownidx.end(), natcmpstratom);

      std::string result;
      result.reserve(102400);
      result += '[';

      for (int i = first, e = shownidx.size(); i < e && i < last; ++i)
      {
	if (i > first)
	  result += ", ";
	result += shownToJSON(contents, status, drawopts,
			      dataset, step, runnr,
			      shownidx[i]);
      }

      result += ']';
      return result;
    }

  static std::string
  shownToJSON(const VisDQMItems &contents,
	      const VisDQMStatusMap &status,
	      const VisDQMDrawOptionMap &drawopts,
	      const std::string &dataset,
	      const std::string &step,
	      const std::string &runnr,
	      const StringAtom &x)
    {
      VisDQMShownRows layout;
      layoutOne(layout, contents, status, drawopts, x);
      return StringFormat("{name: %1, items: %2}")
	.arg(stringToJSON(x.string()))
	.arg(layoutToJSON(layout, dataset, step, runnr));
    }

  static void
  copyAxisOptions(VisDQMAxisOptions &to, const VisDQMAxisOptions &from)
    {
      if (! from.min.empty())
	to.min = from.min;

      if (! from.max.empty())
	to.max = from.max;

      if (! from.type.empty())
	to.type = from.type;
    }

  static void
  copyDrawOptions(VisDQMDrawOptions *to, const VisDQMDrawOptions *from)
    {
      if (! from)
	return;

      if (! from->withref.empty())
	to->withref = from->withref;

      if (! from->drawopts.empty())
	to->drawopts = from->drawopts;

      copyAxisOptions(to->xaxis, from->xaxis);
      copyAxisOptions(to->yaxis, from->yaxis);
      copyAxisOptions(to->zaxis, from->zaxis);
    }

  static void
  setDrawOptions(VisDQMDrawOptions *to,
		 const VisDQMDrawOptions *opt1 = 0,
		 const VisDQMDrawOptions *opt2 = 0)
    {
      to->withref = "no";
      to->drawopts = "";
      to->xaxis.type = "lin";
      to->yaxis.type = "lin";
      to->zaxis.type = "lin";
      copyDrawOptions(to, opt1);
      copyDrawOptions(to, opt2);
    }

  // Build a layout on this node, 1x1 if it's not a layout otherwise.
  static void
  layoutOne(VisDQMShownRows &layout,
	    const VisDQMItems &contents,
	    const VisDQMStatusMap &status,
	    const VisDQMDrawOptionMap &drawopts,
	    const StringAtom &path)
    {
      VisDQMItems::const_iterator          objpos = contents.find(path);
      VisDQMStatusMap::const_iterator      statuspos = status.find(path);
      VisDQMDrawOptionMap::const_iterator  drawpos = drawopts.find(path);

      assert(objpos != contents.end());
      assert(objpos->second->name == path);
      assert(statuspos != status.end());

      const VisDQMItem &obj = *objpos->second;
      const VisDQMStatus &info = statuspos->second;
      const VisDQMDrawOptions *xopts = 0;
      if (drawpos != drawopts.end())
	xopts = &drawpos->second;

      if (! obj.layout)
      {
	shared_ptr<VisDQMShownItem> item(new VisDQMShownItem);
	item->version = obj.version;
	item->nalarm = info.nalarm;
	item->nlive = info.nlive;
	item->name = obj.name;
	item->plotter = obj.plotter;
	setDrawOptions(&item->drawopts, xopts);
	shared_ptr<VisDQMShownRow> row(new VisDQMShownRow);
	row->columns.push_back(item);
	layout.push_back(row);
      }
      else
      {
	const VisDQMLayoutRows &rows = *obj.layout;
	layout.reserve(rows.size());

	for (size_t nrow = 0; nrow < rows.size(); ++nrow)
	{
	  const VisDQMLayoutRow &row = *rows[nrow];
	  shared_ptr<VisDQMShownRow> srow(new VisDQMShownRow);
	  srow->columns.reserve(row.columns.size());
	  layout.push_back(srow);

	  for (size_t ncol = 0; ncol < row.columns.size(); ++ncol)
	  {
	    const VisDQMDrawOptions *lopts = 0;
	    const VisDQMLayoutItem &col = *row.columns[ncol];
	    shared_ptr<VisDQMShownItem> scol(new VisDQMShownItem);

	    if ((objpos = contents.find(col.path)) != contents.end())
	    {
	      statuspos = status.find(col.path);
	      assert(statuspos != status.end());

	      const VisDQMItem &lobj = *objpos->second;
	      const VisDQMStatus &linfo = statuspos->second;
	      assert(lobj.name == col.path);

	      scol->version = lobj.version;
	      scol->nalarm = linfo.nalarm;
	      scol->nlive = linfo.nlive;
	      scol->name = col.path;
	      scol->desc = col.desc;
	      scol->plotter = lobj.plotter;
	      lopts = &col.drawopts;
	    }
	    else
	    {
	      scol->version = 0;
	      scol->nalarm = 0;
	      scol->nlive = 1;
	      scol->name = col.path;
	      scol->desc = col.desc;
	      scol->plotter = 0;
	    }

	    setDrawOptions(&scol->drawopts, lopts, xopts);
	    srow->columns.push_back(scol);
	  }
	}
      }
    }

  static std::string
  treeToJSON(const VisDQMItems &contents,
	     const VisDQMStatusMap &status,
	     const StringAtom &path,
	     const std::string &exclude,
	     const std::string &focus,
	     const std::string &picked,
	     int depth)
    {
      assert(status.count(path));
      const VisDQMStatus &info = status.find(path)->second;
      std::string kidjson;

      if (depth > 0)
      {
	std::string prefix;
	if (! path.string().empty())
	{
	  prefix.reserve(path.size() + 2);
	  prefix += path.string();
	  prefix += '/';
	}

	// Determine sorted list of names immediately below "prefix".
	StringAtomSet kids;
	VisDQMItems::const_iterator ci, ce;
	for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	{
	  const std::string &name = ci->first.string();
	  if (name.size() > prefix.size()
	      && name.compare(0, prefix.size(), prefix) == 0
	      && name.compare(prefix.size(), exclude.size(), exclude) != 0)
	    kids.insert(StringAtom(&stree, std::string(name, 0, name.find('/', prefix.size()))));
	}

	StringAtomList kidnames(kids.begin(), kids.end());
	std::sort(kidnames.begin(), kidnames.end(), natcmpstratom);

	kidjson.reserve(1024 * kidnames.size());
	for (size_t i = 0, e = kidnames.size(); i != e; ++i)
	{
	  if (i > 0)
	    kidjson += ", ";
	  kidjson += treeToJSON(contents, status, kidnames[i],
				exclude, focus, picked, depth-1);
	}
      }

      return StringFormat("{name: %1, nlive: %2, nalarm: %3,"
			  " nleaves: %4, isdir: %5, ispicked: %6, kids: [%7]}")
	.arg(stringToJSON(path.string()))
	.arg(info.nlive)
	.arg(info.nalarm)
	.arg(info.nleaves)
	.arg(info.isdir)
	.arg(int((! focus.empty() && focus == path.string())
		 || (focus.empty() && picked == path.string())))
	.arg(kidjson);
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
#define EXC_TRANSLATE(n,t,what)					\
  struct translate ## n { static void translate (const t &e)	\
    { PyErr_SetString(PyExc_RuntimeError, e. what); } };	\
  boost::python::register_exception_translator< t >		\
    (&translate ## n :: translate)

BOOST_PYTHON_MODULE(pluginVisDQMServer) {
  DebugAids::failHook(&onAssertFail);

  // Re-capture signals from ROOT after ROOT has initialised.
  ROOT::GetROOT();
  for (int sig = 1; sig < _NSIG; ++sig)
    if (sig == SIGPIPE)
      Signal::ignore(sig);
    else
      Signal::revert(sig);

  Signal::handleFatal(0, IOFD_INVALID, 0, 0,
		      (Signal::FATAL_DEFAULT
		       & ~(Signal::FATAL_ON_INT
			   | Signal::FATAL_ON_QUIT
			   | Signal::FATAL_DUMP_CORE)));

  // Register methods.
  py::class_<VisDQMSource, shared_ptr<VisDQMSource>, boost::noncopyable>
    ("DQMSource", py::no_init)
    .def("prepareSession", &VisDQMSource::prepareSession);

  py::class_<VisDQMUnknownSource, shared_ptr<VisDQMUnknownSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMUnknownSource", py::init<>())
    .add_property("plothook", &VisDQMUnknownSource::plotter);

  py::class_<VisDQMLiveSource, shared_ptr<VisDQMLiveSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMLiveSource", py::init<py::object, py::dict>())
    .add_property("plothook", &VisDQMLiveSource::plotter)
    .def("_name", &VisDQMLiveSource::path)
    .def("_stop", &VisDQMLiveSource::stop);

  py::class_<VisDQMArchiveSource, shared_ptr<VisDQMArchiveSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMArchiveSource", py::init<py::object, py::dict>())
    .add_property("plothook", &VisDQMArchiveSource::plotter)
    .def("_name", &VisDQMArchiveSource::path);

  py::class_<VisDQMLayoutSource, shared_ptr<VisDQMLayoutSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMLayoutSource", py::init<py::tuple>());

  py::class_<VisDQMSummaryWorkspace, boost::noncopyable>
    ("DQMSummaryWorkspace", py::init<py::object, std::string>())
    .def("_state", &VisDQMSummaryWorkspace::state);

  py::class_<VisDQMQualityWorkspace, boost::noncopyable>
    ("DQMQualityWorkspace", py::init<py::object, std::string>())
    .def("_state", &VisDQMQualityWorkspace::state);

  py::class_<VisDQMContentWorkspace, boost::noncopyable>
    ("DQMContentWorkspace", py::init<py::object, std::string, std::string>())
    .def("_state", &VisDQMContentWorkspace::state);

  EXC_TRANSLATE(StdError, std::exception, what());
  EXC_TRANSLATE(CMSError, cms::Exception, what());
  EXC_TRANSLATE(LATError, lat::Error, explain().c_str());
}
