#!/usr/bin/env python

# *IMPORTANT NOTE ON THREADS AND LOCKING*: Python currently supports
# multi-threading, but the interpreter is not multi-threaded.  The
# interpreter has a global lock, which it yields on "slow" operations
# such as file and socket operations.  On "straight" python code the
# interpreter yields the lock only every N byte code instructions;
# this server configures a large N (1'000'000).

from copy import deepcopy
from cgi import escape
from socket import gethostname
from threading import Thread, Lock
from cherrypy import expose, HTTPError, request, response, engine, log
from cherrypy.lib.static import serve_file
from Cheetah.Template import Template
from GuiUtils import _logerr, _logwarn, _loginfo
from cStringIO import StringIO
from stat import *
import cPickle as pickle
import sys, os, os.path, re, tempfile, time, inspect, logging, traceback, md5

_SESSION_REDIRECT = ("<html><head><script>location.replace('%s')</script></head>"
                     + "<body><noscript>Please enable JavaScript to use this"
                     + " service</noscript></body></html>")

def extension(modules, what, *args):
  for m in modules:
    c = getattr(m, what, None) 
    if c:
      return c(*args)
  _logwarn("extension '%s' not found" % what)
  return None

# -------------------------------------------------------------------
# Background thread for managing the server's sessions.
#
# Saving each session in the HTTP service threads reduces server
# responsiveness.  For a session modified at a high burst rate in
# asynchronous requests it would be pointless to save every
# modifiation to disk.
#
# The server runs this separate thread to save modified sessions back
# to disk in python's pickled format.  It is essential the per-session
# data is of limited size as pickling is slow.  The output is done in
# as safe a manner as possible to avoid data loss if the server
# crashes or is restarted, and even if it is terminated forcefully
# with SIGKILL.  Modified sessions are written out about once a
# second.
#
# ._stopme   Set by main server thread to indicate it's time to exit.
# ._path     The directory where the sessions are stored.
# ._save     A dictionary of sessions that must be saved.
class SessionThread(Thread):
  def __init__(self, path):
    Thread.__init__(self, name="GUI session thread")
    self._lock = Lock()
    self._stopme = False
    self._path = path
    self._save = {}

  # Record a session for saving.  The session's data in VALUE will
  # become the property of this thread exclusively.
  def save(self, data):
    self._lock.acquire()
    self._save[data['core.name']] = data
    self._lock.release()

  # Tell the thread to stop after it has flushed all the data to disk.
  def stop(self):
    self._lock.acquire()
    self._stopme = True
    self._lock.release()

  # The thread run loop.  Checks for dirty sessions about once a
  # second, and if there are any, grabs the list of currently dirty
  # sessions and starts writing them to disk in pickled format.
  # Manipulates session files as safely as possible.
  def run(self):
    while True:
      self._lock.acquire()
      stopme = self._stopme
      save = self._save
      self._save = {}
      self._lock.release()
      for (name, data) in save.iteritems():
        path = self._path + '/' + name
        tmppath = path + ".tmp"
        f = file(tmppath, "w")
        pickle.dump(data, f)
        f.close()

        try:
          os.remove(path)
        except os.error:
          pass

        try:
          os.rename(tmppath, path)
        except os.error:
          self._lock.acquire()
          self._save[name] = data
          self._lock.release()

      if stopme:
	break
      time.sleep(1)

# -------------------------------------------------------------------
# The main server process, a CherryPy actor mounted to the URL tree.
# The basic server core orchestrates basic services such as session
# management, templates, static content and switching workspaces.
# Behaviour beyond these basic functions is delegated to workspaces;
# all other invocations are delegated to the current workspace of the
# session in question.
#
# Note that all volatile user data is stored in session objects.  The
# backends and worspaces cache data to store global state that is not
# pertinent to any particular user/browser session, allowing them to
# respond HTTP requests as quickly as possible (ideally using only
# data currently in memory).
# 
# Configuration data:
# .title            Web page banner title.
# .baseUrl          URL root for this service.
# .serviceName      Label for this service, to show on the web page.
# .services         List of (label, url) for all monitoring services.
# .workspaces       List of workspaces configured for this service.
# .sources          List of data sources attached to this service.
# .sessiondir       Directory where session data is kept.
#
# Dynamic data or automatically determined configuration:
# .contentpath      Directory where static content will be found.
# .yuipath          Directory for YUI static content files.
# .templates        Cheetah template files in .contentpath.
# .sessions         Currently active sessions.
# .sessionthread    Background thread for saving sessions.
# .lock             Lock for modifying variable data.
# .stamp            Server start time for forcing session reloads.
# .checksums        Server integrity check of source files.
class Server:
  def __init__(self, cfgfile, cfg, modules):
    modules = map(__import__, modules)
    self.checksums = []
    self.stamp = time.time()
    self.lock = Lock()
    self.services = cfg.services
    self.serviceName = cfg.serviceName
    self.templates = {}
    self.contentpath = inspect.getabsfile(self.__class__).rsplit('/', 1)[0]
    self.baseUrl = cfg.baseUrl
    self.sessiondir = cfg.serverDir + '/www/sessions'
    self.title = cfg.title
    self.yuipath = os.getenv("YUI_ROOT") + "/build/"
    for file in os.listdir(self.contentpath):
      m = re.match(r'(.*)\.tmpl$', file)
      if m:
        (base,) = m.groups()
        filename = "%s/%s" % (self.contentpath, file)
        self.templates[base] = [ filename, os.stat(filename)[ST_MTIME], open(filename).read() ]
    self.sessions = {}
    self.sessionthread = SessionThread(self.sessiondir)
    self.extensions = [extension(modules, e[0], self, *e[1])
                       for e in cfg.extensions]
    self.workspaces = [extension(modules, w[0] + "Workspace", self, *w[1])
                       for w in cfg.workspaces]
    self.sources = [extension(modules, s[0] + "Source", self,
                              cfg.serverDir + '/' + s[1], *s[2])
                    for s in cfg.sources]

    self._addChecksum(None, cfgfile, open(cfgfile).read())
    for name, m in sys.modules.iteritems():
      if (name.startswith("Gui") or name == "__main__") \
	 and m and m.__dict__.has_key('__file__'):
        self._addChecksum(name,
			  inspect.getsourcefile(m) \
			  or inspect.getabsfile(m)
			  or name,
			  inspect.getsource(m))

    self.sessionthread.start()
    engine.on_stop_engine_list.append(self.sessionthread.stop)

  def _addChecksum(self, modulename, file, data):
    s = (os.path.exists(file) and os.stat(file)) or None
    self.checksums.append({
      'module': modulename,
      'srcfile': file,
      'srcname': file.rsplit("/", 1)[-1],
      'mtime': (s and s[ST_MTIME]) or None,
      'srclen': (s and s[ST_SIZE]) or None,
      'srcmd5': md5.new(data).digest().encode('hex')
    })

  def _maybeRefreshFile(self, dict, name):
    self.lock.acquire()
    fileinfo = dict[name]
    mtime = os.stat(fileinfo[0])[ST_MTIME]
    if mtime != fileinfo[1]:
      fileinfo[1] = mtime
      fileinfo[2] = open(fileinfo[0]).read()
    self.lock.release()
    return fileinfo[2]
 
  def _templatePage(self, name, variables):
    template = self._maybeRefreshFile(self.templates, name)
    return str(Template(template, searchList=[variables]))

  def _noResponseCaching(self):
    response.headers['Pragma'] = 'no-cache'
    response.headers['Expires'] = 'Sun, 19 Nov 1978 05:00:00 GMT'
    response.headers['Cache-Control'] = \
      'no-store, no-cache, must-revalidate, post-check=0, pre-check=0'

  # -----------------------------------------------------------------
  # Session methods.
  def _sessionClientData(self):
    return request.remote.ip

  # Check NAME is valid and a known session.  If yes, returns the
  # session data, otherwise None.  Locks the session before returning
  # it, making sure all other threads have released the session.  The
  # caller _MUST_ release the session lock before the HTTP request
  # handling returns, or the next access to the session will hang.
  def _getSession(self, name):
    s = None
    if re.match("^[-A-Za-z0-9_]+$", name):
      self.lock.acquire()
      if name not in self.sessions:
	path = "%s/%s" % (self.sessiondir, name)
        if os.path.exists(path):
	  try:
            self.sessions[name] = pickle.load(file(path, "r"))
	  except Exception, e:
	    _logerr("FAILURE: cannot load session data: " + str(e))

      s = self.sessions.get(name, None)
      self.lock.release()

      if s:
        current = self._sessionClientData()
        if not isinstance(s['core.clientid'], str):
	  s['core.clientid'] = s['core.clientid']['requestip']
        if s['core.clientid'] == current and s['core.name'] == name:
          if 'core.lock' not in s:
            s['core.lock'] = Lock()
          s['core.lock'].acquire()
	else:
	  s = None
    return s

  # Save the SESSION state.
  def _saveSession(self, session):
    self.lock.acquire()
    self.sessions[session['core.name']] = session
    self.lock.release()
    self.sessionthread.save(dict((k, deepcopy(session[k]))
                                 for k in session.keys()
                                 if k != 'core.lock'))

  # Release the SESSION for use by other threads.
  def _releaseSession(self, session):
    session['core.lock'].release()

  # Create and initialise a new session with some default workspace.
  # This just initialises a session; it will not become locked.
  def _newSession(self, workspace):
    # Generate a new session key.
    (fd, path) = tempfile.mkstemp("", "", self.sessiondir)
    sessionid = (path.split('/'))[-1]
    os.close(fd)

    # Build session data.  We record:
    #  - client data for later verification
    #  - which workspace we are in
    session = {}
    session['core.name'] = sessionid
    session['core.clientid'] = self._sessionClientData()
    session['core.public'] = False
    session['core.workspace'] = self.workspaces[0].name

    # Get default objects in the workspace.
    workspace = workspace.lower()
    for w in self.workspaces:
      if w.name.lower() == workspace:
        session['core.workspace'] = w.name
      w.initialiseSession(session)

    # Get defaults also from the sources.
    for s in self.sources:
      s.prepareSession(session)

    # Save the pickled session state file.
    self._saveSession(session)

    # Return the final component of the session path.
    return sessionid

  # Tell the client browser the URL is invalid.
  def _invalidURL(self):
    return self._templatePage("invalid", {
	'TITLE'		 : re.sub(r"\&\#821[12];", "-", self.title),
	'HEADING'	 : self.title,
	'URL'		 : escape(request.request_line.split(' ')[1]),
	'ROOTPATH'	 : self.baseUrl,
	'NEWSESSION'	 : self.baseUrl,
	'HOSTNAME'	 : gethostname(),
      })

  # Get the workspace object corresponding to NAME.  If no such
  # workspace exists, returns the first (= default) workspace.
  def _workspace(self, name):
    name = name.lower()
    for w in self.workspaces:
      if w.name.lower() == name:
	return w
    return self.workspaces[0]
    
  # -----------------------------------------------------------------
  # Server access points.

  # Main root index address: the landing address.  Create a new
  # session and redirect the client there.
  @expose
  def index(self):
    return self.start(workspace = self.workspaces[0].name);

  # Access YUI static content.
  @expose
  def YUI(self, *args, **kwargs):
    if ("." not in args) and (".." not in args):
      return serve_file(self.yuipath + "/".join(args))

  # Access our own static content.
  @expose
  def static(self, *args, **kwargs):
    if len(args) != 1 or not re.match(r"^[a-z]+\.(js|css|png|gif)$", args[0]):
      return self._invalidURL()
    return serve_file(self.contentpath + '/' + args[0])

  # -----------------------------------------------------------------
  # Jump to some content.  This creates and configures a new session with
  # the desired content, as if a sequence of actions was carried out.
  #
  # In the end, redirects the client browser to a new session URL.  We
  # send a HTML page with JavaScript to change the page "location",
  # rather than raise a HTTPRedirect.  The main reason is HTTPRedirect
  # results in HTTP 303 response and web browsers remember the
  # original address not the new one we send to them.  If the user
  # would then reload the page, they would not be sent back to their
  # session but back to the root address which would create them again
  # another session.  The second and minor reason is that we can
  # verify that JavaScript is enabled in the client browser.
  @expose
  def start(self, *args, **kwargs):
    if len(args) != 0:
      return self._invalidURL()
    workspace = self._workspace(kwargs.get("workspace", self.workspaces[0].name))
    sessionid = self._newSession(workspace.name)
    session = self._getSession(sessionid)
    workspace.start(session, **kwargs)
    self._saveSession(session)
    self._releaseSession(session)
    return _SESSION_REDIRECT % (self.baseUrl + "/session/" + sessionid)

  # Backward compatible version of 'start' which understands one
  # parameter, the name of the workspace to begin in.  Note that
  # within a session the sessionWorkspace method is used to switch
  # between workspaces.
  @expose
  def workspace(self, *args, **kwargs):
    if len(args) == 1:
      return self.start(workspace = self._workspace(args[0]).name)
    else:
      return self.start(workspace = self.workspaces[0].name)

  # -----------------------------------------------------------------
  # General session-independent access path for dynamic images.
  # The first subdirectory argument contains the name of the
  # "source plot hook" able to handle the plotting request.
  # The rest of the processing is given over to the hook.
  @expose
  def plotfairy(self, *args, **kwargs):
    try:
      if len(args) >= 1:
        for s in self.sources:
          if getattr(s, 'plothook', None) == args[0]:
            (type, data) = s.plot(*args[1:], **kwargs)
            if type != None:
              self._noResponseCaching()
              response.headers['Content-Length'] = str(len(data))
              response.headers['Content-Type'] = type
              return data
            break
    except Exception, e:
      o = StringIO()
      traceback.print_exc(file=o)
      log("WARNING: unable to produce a plot: "
          + (str(e) + "\n" + o.getvalue()).replace("\n", " ~~ "),
          severity=logging.WARNING)

    self._noResponseCaching()
    return serve_file(self.contentpath + "/missing.png",
                      content_type = "image/png")

  # -----------------------------------------------------------------
  # Report code running in this server.
  @expose
  def digest(self, *args, **kwargs):
    maxsize = max(len(str(i['srclen'])) for i in self.checksums) + 1
    maxlen = max(len(str(i['module'])) for i in self.checksums) + 2
    fmt = "%(srclen)-" + str(maxsize) \
	  + "d %(mtime)-10d  %(srcmd5)s  %(module)-" + str(maxlen) \
          + "s %(srcname)s\n"
    summary = ""

    self.checksums.sort(lambda a, b: cmp(a['srcfile'], b['srcfile']))
    for i in self.checksums:
      summary += fmt % i

    response.headers['Content-Type'] = "text/plain"
    return summary

  # -----------------------------------------------------------------
  # Main session address.  All AJAX calls to the session land here.
  # The URL is of the form "[/ROOT]/session/ID[/METHOD].  We check
  # the session ID is valid for this user, and the METHOD is one we
  # support.  A METHOD "foo" results in call to "sessionFoo()" in
  # this class.  If no METHOD is given, default to "index": generate
  # the main GUI page. All other METHODs are AJAX calls from the
  # client, which normally just return a JSON result object.
  @expose
  def session(self, *args, **kwargs):
    # If the URL has been truncated, just start a new session.
    if len(args) < 1:
      return self.start(workspace = self.workspaces[0].name)

    # If the URL specifies extra arguments, reply with an error page.
    sessionid = args[0]
    if len(args) > 2:
      return self._invalidURL()

    # Check the session is something we can use.  This locks the
    # session so no other thread can access it at the same time.
    session = self._getSession(sessionid)
    if not session:
      return self._invalidURL()

    # Find the session method to invoke and verify we have a good one.
    # Look for the method first in this server object, and failing
    # that in the current workspace object.  Release the session lock
    # on the way out.
    try:
      methodName = 'sessionIndex'
      if len(args) == 2:
        methodName = 'session' + args[1][0].upper() + args[1][1:]

      workspace = session['core.workspace']
      method = getattr(self, methodName, None)
      if not method and workspace:
        method = getattr(self._workspace(workspace), methodName, None)
      if not method:
        return self._invalidURL()

      # Let the session method handle the rest.
      return method(session, *args, **kwargs)
    finally:
      self._releaseSession(session)

  # Generate top level session index.  This produces the main GUI web
  # page, with practically no content in it; the client will contact
  # us for the content using AJAX calls.
  def sessionIndex(self, session, *args ,**kwargs):
    return self._templatePage("index", {
	'TITLE'		 : re.sub(r"\&\#821[12];", "-", self.title),
	'HEADING'	 : self.title,
        'WORKSPACE'	 : session['core.workspace'],
        'SESSION_ID'     : session['core.name'],
        'SESSION_STATUS' : 'modifiable',
        'HOSTNAME'       : gethostname(),
	'ROOTPATH'	 : self.baseUrl
      });

  # Switch session to another workspace.
  def sessionWorkspace(self, session, *args, **kwargs):
    workspace = self._workspace(kwargs.get('name', session['core.workspace']))
    session['core.workspace'] = workspace.name
    workspace.initialiseSession(session)
    self._saveSession(session)
    return workspace.sessionState(session)
