#!/usr/bin/env python

from stat import *
from copy import deepcopy, copy
from threading import Thread, Lock
from pysqlite2 import dbapi2 as db
from GuiUtils import _logerr, _logwarn, _loginfo, natsorted, thousands
from cherrypy import expose, HTTPError, tree, response, engine, thread_data, log
from cherrypy.lib.static import serve_file
from struct import pack, unpack
from cStringIO import StringIO
import pluginVisDQMServer as Accelerator
import os, re, time, socket, shutil, tempfile
try:
  from PIL import Image, ImageEnhance
  HAVE_PIL=True
except:
  HAVE_PIL=False

# --------------------------------------------------------------------
# Validate common DQM image request options.
RX_OPT_FLOAT = re.compile(r"^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)?$")
def _checkCommonImageOptions(options):
  return (options.get('w', '').isdigit()
          and options.get('h', '').isdigit()
          and options.get('xtype', '') in ('', 'lin', 'log')
          and re.match(RX_OPT_FLOAT, options.get('xmin', ''))
          and re.match(RX_OPT_FLOAT, options.get('xmax', ''))
          and options.get('ytype', '') in ('', 'lin', 'log')
          and re.match(RX_OPT_FLOAT, options.get('ymin', ''))
          and re.match(RX_OPT_FLOAT, options.get('ymax', ''))
          and options.get('ytype', '') in ('', 'lin', 'log')
          and re.match(RX_OPT_FLOAT, options.get('ymin', ''))
          and re.match(RX_OPT_FLOAT, options.get('ymax', ''))
          and re.match(r"^[A-Za-z ]*$", options.get('drawopts', ''))
          and options.get('withref', '') in ('', 'yes', 'no'))

def ROOTImage(sockname, sockargs, mtype, rtype, options, path, *args):
  if not _checkCommonImageOptions(options):
    return (None, None)

  spec = ""
  resize = None
  MIN_W, MIN_H = 532, 400
  for opt in ('drawopts', 'withref',
              'xtype', 'xmin', 'xmax',
              'ytype', 'ymin', 'ymax',
              'ztype', 'zmin', 'zmax'):
    if options.get(opt, '') != '':
      spec += "%s%s=%s" % ((spec and ";") or "", opt, options[opt])

  # If we have PIL and the image is small, generate a larger image
  # in the same aspect ratio and use PIL to produce a reduced size
  # version.  ROOT generates badly anti-aliased output otherwise.
  if HAVE_PIL and 'w' in options and 'h' in options \
     and options['w'].isdigit() and options['h'].isdigit():
    w, h = int(options['w']), int(options['h'])
    if w > 0 and h > 0 and (w < MIN_W or h < MIN_H):
      resize = (w, h)
      aspect = float(w)/float(h)
      w = max(w, MIN_W)
      spec += "%sw=%d;h=%d" % ((spec and ";") or "", w, w/aspect)

  if not resize:
    for opt in ('w', 'h'):
      if options.get(opt, '') != '':
        spec += "%s%s=%s" % ((spec and ";") or "", opt, options[opt])

  # If we have PIL and the image is small, generate a larger image
  # in the same aspect ratio and use PIL to produce a reduced size
  # version.  ROOT generates badly anti-aliased output otherwise.

  # Now try to get the image from the backend.  Make a few tries
  # in case the socket has become invalid, or the backend dies
  # while we try.
  tryit = 3
  imgtype = None
  imgdata = None
  while tryit > 0:
    tryit -= 1
    try:
      s = getattr(thread_data, sockname, None)
      if not s:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(sockargs)
        setattr(thread_data, sockname, s)

      payload = [path, spec] + [i for i in args]
      paylen = [len(i) for i in payload]
      nlen = 8 + 4 * len(paylen) + sum(paylen)
      payfmt = '%dI' % len(payload)
      s.send(pack('2I', nlen, mtype))
      s.send(pack(payfmt, *paylen))
      s.send("".join(i for i in payload))

      reply = ""
      while s:
        part = s.recv(1048576)
	tryit = 0
        if len(part) == 0:
          s = None
          setattr(thread_data, sockname, s)

        reply += part
        if len(reply) < 8:
          continue

        (length, type) = unpack('2I', reply[0:8])
        if len(reply) < length:
          continue

        pos = 8
        nlen = 0
        datapos = 8 + (len(payload)+1)*4

        assert(len(reply) == length)
        assert(type == rtype)

        for i in xrange(0, len(payload)):
          nlen, = unpack('I', reply[pos:pos+4])
          part = reply[datapos:datapos+nlen]
          assert(part == payload[i])
          datapos += nlen
          pos += 4

        datalen, = unpack('I', reply[pos:pos+4])
        if datalen > 0:
          imgdata = reply[datapos:]
          imgtype = "image/png"
	  break

    except Exception, e:
      _logerr("failed to request image '%s' as %s %s: %s"
	      % (path, spec, args, str(e)))
      setattr(thread_data, sockname, None)

  if HAVE_PIL and imgtype and resize:
    try:
      out = StringIO()
      ImageEnhance.Sharpness(Image.open(StringIO(imgdata))
			     .resize(resize, Image.ANTIALIAS)) \
	.enhance(1.5).save(out, "png")
      imgdata = out.getvalue()
    except Exception, e:
      _logerr("failed to down scale image: %s" % e)

  return (imgtype, imgdata)

# --------------------------------------------------------------------
class DQMItem:
  def __init__(self, layout=None):
    self.live = 0
    self.version = 0
    self.alarm = 0
    self.layout = layout
    self.data = None
    self.flags = 0
    self.origin = None
    self.plotter = None

# --------------------------------------------------------------------
# Fake DQM backend which manages DQM Data file uploads and downloads.
class DQMFileAccess:
  STATUS_OK               = 100 # Requested operation succeeded
  STATUS_BAD_REQUEST      = 200 # The request is malformed
  STATUS_ERROR_PARAMETER  = 300 # Request parameter value is unacceptable.
  STATUS_ERROR_EXISTS     = 301 # Cannot overwrite an existing object.
  STATUS_ERROR_NOT_EXISTS = 302 # Requested file does not exist.
  STATUS_FAIL_EXECUTE     = 400 # Failed to execute the request.

  def refresh(self, *args): pass
  def __init__(self, server, aclfile, datadir):
    self.lock = Lock()
    self.server = server
    self.datadir = datadir
    if not os.path.exists(datadir):
      os.makedirs(datadir)
    tree.mount(self, script_name=server.baseUrl + "/data",
               config={"/": {'request.show_tracebacks': False}})

  # Set response headers to indicate our status data.
  def _status(self, code, message, detail=None):
    response.headers['dqm-status-code'] = str(code)
    response.headers['dqm-status-message'] = message
    response.headers['dqm-status-detail'] = detail

  # Set response headers to indicate an error, then get out.
  def _error(self, code, message, detail=None):
    _logerr("code=%d, message=%s, detail=%s" % (code, message, detail))
    self._status(code, message, detail)
    raise HTTPError(500, message)

  # Check that a required parameter has been given just once,
  # and the value matches the given regular expression.
  def _check(self, name, arg, rx):
    if not arg or not isinstance(arg, str):
      self._error(self.STATUS_BAD_REQUEST,
                  "Incorrect or missing %s parameter" % name,
                  "Must provide single argument")
    if not re.match(rx, arg):
      self._error(self.STATUS_BAD_REQUEST,
                  "Malformed %s argument" % name,
                  "Argument must match regular expression '%s'" % rx)

  # ------------------------------------------------------------------
  # Store a file to the server.  Validates all the parameters, then
  # attempts to save the file safely.  Sets headers in the response
  # to indicate what happened, either an error or success.
  # FIXME: Verify the request was submitted securely and is permitted.
  # FIXME: Determine producer from certificate information.
  @expose
  def put(self,
          producer = None,
          step = None,
          workflow = None,
          filename = None,
          mssname = None,
          size = None,
          checksum = None,
          file = None,
          *args,
          **kwargs):
    # Basic argument validation.
    if file == None \
       or not getattr(file, "file", None) \
       or not getattr(file, "filename", None):
      self._error(self.STATUS_BAD_REQUEST,
                  "Incorrect or missing file argument",
                  "Must provide a single file-type argument")

    self._check("producer", producer,      r"^[A-Za-z0-9]+$")
    self._check("step",     step,          r"^[-A-Za-z0-9]+$")
    self._check("workflow", workflow,      r"^(/[-A-Za-z0-9_]+){3}$")
    self._check("filename", file.filename, r"^[-A-Za-z0-9_]+\.root$")
    self._check("mssname",  mssname,       r"^[-A-Za-z0-9_./:=?]+$")
    self._check("size",     size,          r"^\d+$")
    self._check("checksum", checksum,      r"^(md5:[A-Za-z0-9]+|crc32:\d+)$")

    # Argument value validation.
    if producer not in ("ProdSys",):
      self._error(self.STATUS_ERROR_PARAMETER,
                  "This producer is not authorised")
    if step not in ("Pass-1", "PromptReco"):
      self._error(self.STATUS_ERROR_PARAMETER,
                  "This step is not authorised")

    m = re.match(r"^(DQM)_V\d+_R(\d+)(__.*)\.root", file.filename)
    if not m:
      self._error(self.STATUS_ERROR_PARAMETER,
                  "File name does not match the expected convention")

    size = int(size)
    runnr = int(m.groups()[1])
    wflow = m.groups()[2]
    if workflow.replace("/", "__") != wflow:
      self._error(self.STATUS_ERROR_PARAMETER,
                  "File name does not match the workflow name")
    stream = workflow.split("/")[1]

    # Determine where we would save this file.
    dir = "%s/%s/%s/%s/%03d/%03d/%03d" % \
          (self.datadir, producer, step, stream,
           (runnr / 1000000) % 1000, (runnr / 1000) % 1000, runnr % 1000)
    fname = "%s/%s" % (dir, file.filename)
    tmp = None

    if fname.find("/../") >= 0 or fname.find("/./") >= 0:
      self._error(self.STATUS_ERROR_PARAMETER, "Invalid file path name",
                  "Path name must not refer to '.' or '..'")

    if mssname.find("/../") >= 0 or mssname.find("/./") >= 0:
      self._error(self.STATUS_ERROR_PARAMETER, "Invalid file path name",
                  "Path name must not refer to '.' or '..'")

    if os.path.exists(fname) or os.path.exists(fname + ".origin"):
      self._error(self.STATUS_ERROR_EXISTS,
                  "File exists, cannot overwrite", fname)

    # Try saving the file safely.  First we write the file to a
    # temporary directory, with a fake name, to the same directory as
    # the final file would be.  We then move it in place atomically
    # and leave a marker to indicate where the MSS replica of the file
    # can be found. If anything goes wrong, clean up so the upload can
    # be re-attempted later without tripping over the checks above.
    try:
      self.lock.acquire()
      if not os.path.exists(dir):
        os.makedirs(dir)
      (fd, tmp) = tempfile.mkstemp(".upload", "", dir)
      nsaved = 0
      while True:
        data = file.file.read(32*1024)
        if not data:
          break
        os.write(fd, data)
        nsaved += len(data)
      os.close(fd)
      os.chmod(tmp, 0644)
      if nsaved != size:
        self._error(self.STATUS_FAIL_EXECUTE,
                    "Failed to save file data",
                    "Wrote %d bytes, expected to write %d" % (nsaved, size))
      shutil.move(tmp, fname)
      open(fname + ".origin", "w").write("%s %d %s\n" % (checksum, size, mssname))
      self.lock.release()

    except Exception, e:
      self.lock.release()
      if os.path.exists(fname):
        os.remove(fname)
      if os.path.exists(fname + ".origin"):
        os.remove(fname + ".origin")
      if tmp and os.path.exists(tmp):
        os.remove(tmp)
      if isinstance(e, HTTPError):
        raise e
      self._error(self.STATUS_FAIL_EXECUTE,
                  "Failed to save file data",
                  str(e).replace("\n", "; "))

    # Indicate success.
    self._status(self.STATUS_OK, "File saved",
                 "Wrote %d bytes, reference in %s" % (nsaved, mssname))
    _loginfo("saved file path=%s msspath=%s size=%d mssname=%s"
             % (fname, tmp, size, mssname))
    return "Thanks.\n"

  # ------------------------------------------------------------------
  # Default path handler to retrieve a file from the server.  Serves
  # the file if the path is a valid and safe reference to an archived
  # file.  (FIXME: Limiting bandwidth?)
  @expose
  def default(self, *args, **kwargs):
    if len(args) == 0:
      self._error(self.STATUS_BAD_REQUEST, "Must provide file path")
    path = "/".join(args)
    self._check("path", path, r"^[A-Z][-A-Za-z0-9_./]+\.root$")
    if path.find("/../") >= 0 or path.find("/./") >= 0:
      self._error(self.STATUS_ERROR_PARAMETER,
                  "Invalid path name",
                  "Path name must not refer to '.' or '..'")
    fname = "%s/%s" % (self.datadir, path)
    _loginfo("download file %s" % fname)

    if not os.path.exists(fname) or not os.path.isfile(fname):
      self._error(self.STATUS_ERROR_NOT_EXISTS,
                  "File does not exist", path)

    self._status(self.STATUS_OK, "Transferring file", str(os.stat(fname)[ST_SIZE]))
    return serve_file(fname, content_type="application/octet-stream")

# --------------------------------------------------------------------
# Source for plotting non-existent DQM objects, using the archive source.
class DQMUnknownSource(Accelerator.DQMUnknownSource):
  # Notify plot fairy the subdirectory where we serve images.
  type = "DQMArchive"
  port = 9092

  def __init__(self, server, statedir, type = None, port = None):
    Accelerator.DQMUnknownSource.__init__(self)
    if type: self.type = type
    if port: self.port = port

  # Generate an image for an unknown object.
  def plot(self, run, dsPrim, dsWflow, dsTier, step, *path, **options):
    pathname = "/".join(path)
    if self.type == "DQMArchive":
      return ROOTImage('archreqsocket', ("localhost", self.port),
		       5, 106, options, pathname, "")
    elif self.type == "DQMLive":
      return ROOTImage('livereqsocket', ("localhost", self.port),
		       4, 105, options, pathname)
    else:
      return (None, None)

# --------------------------------------------------------------------
# DQM data source which provides content from python configuration
# files.  There is no backend process attached to this data source.
#
# Takes a list of python configuration filen ames as the configuration
# parameter.  The files contain code to initialise a static structure
# of layout definitions.
#
# When ever the source is prompted to update its data, it first
# reviews the list of layout configuration files.  Changed files are
# first reloaded; if the file fails to load it is skipped, although
# layout registrations up to the point of error are included.
#
# Once the layout definitions are up-to-date, simply returns the list
# of known layouts.  No checks are made to ensure the listed monitor
# elements actually exist; non-existent ones will simply generate a
# warning when rendered.
#
# This source does not produce plots.  Plots are provided by the
# sources which provide the objects the layouts refer to.
#
# .lastupdate   The last modification time of each configuration file.
# .filenames    The list of python configuration files.
# .items        The known layout monitor elements.
# .lock         Lock to protect against race conditions.
class DQMLayoutSource(Accelerator.DQMLayoutSource):
  def __init__(self, server, statedir, *filenames):
    Accelerator.DQMLayoutSource.__init__(self, filenames)
    for f in filenames:
      server._addChecksum(None, f, open(f).read())

# --------------------------------------------------------------------
# DQM data source providing content from DQM data delivered over
# network.  Data is received directly from upstream "DQMCollector"
# and rendered by an auxiliary "visDQMServe" process.
#
# Takes as parameter the options for the "visDQMServe" process.  This
# GUI server side object decodes those arguments and determines the
# host and port to connect to reach the backend processes.
#
# All the monitoring and processing is done in the C++ accelerator
# class this python class inherits from.  The python layer is just
# a glue layer to pass data between the two.
#
# Image requests are sent to the backend process via a socket.  The
# request blocks until the backend returns the image data.  Each HTTP
# server thread has a semi-persistent socket to the render process,
# cf. CherryPy's thread_data.  Note that there is no time out on
# retrieving the image data from the backend; the backend process will
# automatically time out after a while, and it will commit a suicide
# if it gets stuck, and the browser/upstream proxy side will time out
# eventually as well.
#
# .plothook     The subdirectory where we appear for plotting.
# .opts         Setting parameters to the native implementation (mostly).
# 
# thread_data.livereqsocket
#               Per server thread socket to the backend for image requests.
#               Currently not in use, a per image request socket is used.
class DQMLiveSource(Accelerator.DQMLiveSource):
  # Start off the live backend and hook up to server transitions.
  def __init__(self, server, statedir, *params):
    self.opts = { "dataset": "/Global/Online/ALL",
                  "step": "Global run",
                  "runnr": "Live",
                  "verbose": False,
                  "dqmhost": "localhost",
                  "dqmport": 9090,
                  "plothost": "localhost",
                  "plotport": 9091 }
    for arg in params:
      m = re.match(r"--collector ((\S+):)?(\d+)", arg)
      if m:
        if m.group(2) != "":
          self.opts['dqmhost'] = m.group(2)
        self.opts['dqmport'] = int(m.group(3))
        continue

      m = re.match(r"--listen (\d+)", arg)
      if m:
        self.opts['plotport'] = int(m.group(1))
        continue

    Accelerator.DQMLiveSource.__init__(self, server, self.opts)
    engine.on_stop_engine_list.append(self._stop)
    engine.on_start_thread_list.append(self._startThread)
    engine.on_stop_thread_list.append(self._stopThread)

  # Routine invoked when a new HTTP server thread starts.  Initialises
  # a per-thread socket variable, used for persistent connections for
  # image requests to the backend, but does not yet start the
  # connection.
  def _startThread(self, id):
    thread_data.livereqsocket = None

  # Routine invoked when a HTTP server thread quits.  Removes the
  # per-thread image request socket if there is one.
  def _stopThread(self, id):
    thread_data.livereqsocket = None

  # Generate an image for an object.  Validates image request options
  # and builds a spec string to obtain the image from the rendering
  # backend process.  Talks to the backend process to obtain an image.
  #
  # Returns string tuple (ImageMimeType, ImageData) to the caller if
  # the object was successfully rendered.  Returns (None, None) if the
  # render request cannot be fulfilled for any reason, be that due to
  # parameter validation issues or real problems.
  #
  # The render backend process handles timeout behaviour.  So we may
  # build up 50 or HTTP server threads hung up waiting for an image to
  # come here, preventing the process from serving any other requests.
  #
  # The code uses a single socket per HTTP request thread to the
  # backend rendering process to avoid excess per-request costs.
  def plot(self, run, dsPrim, dsWflow, dsTier, step, *path, **options):
    n = self._name("/".join(('', dsPrim, dsWflow, dsTier)), step, run, "/".join(path))
    if n != "":
      return ROOTImage('livereqsocket',
                       (self.opts["plothost"], self.opts["plotport"]),
		       4, 105, options, n)
    else:
      return (None, None)

# --------------------------------------------------------------------
# DQM data source providing content from archived DQM data files.  The
# available contents is determined solely from an SQLite database
# maintained by an outside cron-scheduled process.  A backend process
# running "visDQMFile" does the actual histogram rendering, similar to
# the "visDQMServe" for live DQM data.
#
# Takes as parameters the location of the "master" SQLite database
# and options to the "visDQMFile" process.  This GUI server side
# object decodes the backend arguments to determine the host and
# port to connect to reach the backend.
#
# The data source maintains its own private copy of the SQLite
# database in the web server state directory.  Every once in a while
# it checks to see if the master database has changed, and if so,
# copies it over again.  This is so that the master can be modified by
# the cron-scheduled task to update the contents without locking
# coordination with the web server.
#
# Note that the web server uses the SQLite database in strictly
# read-only mode, it will never modify any data in it.
#
# The SQLite database contains all the information necessary to
# represent the object tree and the attributes of each object.  The
# render requests to the backend work very much the same as with the
# DQMLiveSource, including how the parameters are communicated.  This
# source adds the information on which ROOT file to use and where to
# locate the requested monitor element in that file.
#
# .plothook     The subdirectory where we appear for plotting.
# .host         The backend server process host to connect to.
# .port         The backend server process port to connect to.
# .master       The location of the master file.
# .local        The location of the local file.
# .lastupdate   The time the data was last read in.
# .db           Handle to access the database file.
# .mdatasets    Mapping table of dataset ids to names.
# .mstepss      Mapping table of step ids to names.
# .msystems     Mapping table of system ids to names.
# .mcategories  Mapping table of category ids to names.
# .mfiles       Mapping table of files ids to file names.
# .mnames       Mapping table of name ids to names.
# .runs         List of all known runs.
# .data         The actual monitoring object data in compact form.
# .recent	Time stamps when runs were most recently accessed.
# .rundata	Precomputed run info (run, lumi, event) per run.
# .lock         Lock to protect against race conditions.
#
# thread_data.archreqsocket
#               Per server thread socket to the backend for image requests.
#               Currently not in use, a per image request socket is used.
class DQMArchiveSource(Accelerator.DQMArchiveSource):
  # Notify plot fairy the subdirectory where we serve images.

  def __init__(self, server, statedir, dbfile, *params):
    self.opts = { "master": dbfile,
                  "local": statedir + "/dqmfiles.db",
                  "plothost": "localhost",
                  "plotport": 9092 }
    for arg in params:
      m = re.match(r"--listen (\d+)", arg)
      if m:
        self.opts['plotport'] = int(m.group(1))
        continue

    Accelerator.DQMArchiveSource.__init__(self, server, self.opts)
    engine.on_start_thread_list.append(self._startThread)
    engine.on_stop_thread_list.append(self._stopThread)

  # Routine invoked when a new HTTP server thread starts.  Initialises
  # a per-thread socket variable, used for persistent connections for
  # image requests to the backend, but does not yet start the
  # connection.
  def _startThread(self, id):
    thread_data.archreqsocket = None

  # Routine invoked when a HTTP server thread quits.  Removes the
  # per-thread image request socket if there is one.
  def _stopThread(self, id):
    thread_data.archreqsocket = None

  # Generate an image for an object.  Validates image request options
  # and builds a spec string to obtain the image from the rendering
  # backend process.  Talks to the backend process to obtain an image.
  #
  # Returns string tuple (ImageMimeType, ImageData) to the caller if
  # the object was successfully rendered.  Returns (None, None) if the
  # render request cannot be fulfilled for any reason, be that due to
  # parameter validation issues or real problems.
  #
  # The render backend process handles timeout behaviour.  So we may
  # build up 50 or HTTP server threads hung up waiting for an image to
  # come here, preventing the process from serving any other requests.
  #
  # The code uses a single socket per HTTP request thread to the
  # backend rendering process to avoid excess per-request costs.
  def plot(self, run, dsPrim, dsWflow, dsTier, step, *path, **options):
    fname, name = self._name("/".join(('', dsPrim, dsWflow, dsTier)),
			     step, run, "/".join(path))
    if fname and name:
      return ROOTImage('archreqsocket',
                       (self.opts["plothost"], self.opts["plotport"]),
		       5, 106, options, name, fname)
    else:
      return (None, None)

# --------------------------------------------------------------------
# .sessiondef
# .gui
# .rank
# .category
# .name
# .match
# .warnings         Recent warnings printed to the logs.

class DQMWorkspace:
  sessiondef = { 'dqm.dataset':  "",
                 'dqm.step':     "",
                 'dqm.runnr':    "",
		 'dqm.qplot':	 "",
                 'dqm.search':   "",
                 'dqm.filter':   "all",
                 'dqm.size':     "M",
                 'dqm.playmode': 0,
                 'dqm.playinterval': 5,
                 'dqm.playpos':  0,
                 'dqm.perpage':  12,
                 'dqm.root':     {},
                 'dqm.picked':   {},
                 'dqm.focus':    {},
                 'dqm.page':     {},
                 'dqm.drawopts': {},
                 'dqm.myobjs':   {} }

  def __init__(self, gui, rank, category, name):
    self.gui       = gui
    self.rank      = rank
    self.category  = category
    self.name      = name
    self.warnings  = {}

  # Initialise a new session.
  def initialiseSession(self, session):
    for (var, value) in self.sessiondef.iteritems():
      if var not in session:
        session[var] = deepcopy(value)

  # Make sure the accessed object name is valid.
  def _checkObjectName(self, o):
    if not re.match(r"^[-+=_()/# A-Za-z0-9]*$", o):
      if o not in self.warnings:
        _logwarn("attempt to access unsafe object or version '%s'" % o)
	self.warnings[o] = 1
      raise HTTPError(500, "Attempt to access unsafe object")
    return o

  # -----------------------------------------------------------------
  # Initialise a new session with the "start" URL.
  def start(self, session,
            dataset = None,
            step = None,
            runnr = None,
	    qplot = None,
            filter = None,
            search = None,
            size = None,
            perpage = None,
            playmode = None,
            playinterval = None,
            playpos = None,
            root = None,
            picked = None,
            focus = None,
            page = None,
            reset = None,
            add = None,
            **kwargs):

    if dataset != None:
      if not isinstance(dataset, str):
        raise HTTPError(500, "Incorrect dataset parameter")
      session['dqm.dataset'] = dataset

    if step != None:
      if not isinstance(step, str):
        raise HTTPError(500, "Incorrect step parameter")
      session['dqm.step'] = step

    if qplot != None:
      if not isinstance(qplot, str):
        raise HTTPError(500, "Incorrect quality plot parameter")
      if not re.match(r"^[A-Za-z.]*$", qplot):
        raise HTTPError(500, "Invalid quality plot parameter")
      session['dqm.qplot'] = qplot

    if runnr != None:
      if not isinstance(runnr, str):
        raise HTTPError(500, "Incorrect run parameter")
      elif runnr == "Live" or runnr == "":
        session['dqm.runnr'] = "Live"
      elif not re.match(r"^\d+$", runnr):
        raise HTTPError(500, "Incorrect run parameter")
      else:
        session['dqm.runnr'] = runnr

    if filter != None:
      if not isinstance(filter, str) \
         or filter not in ("all", "alarms", "nonalarms"):
        raise HTTPError(500, "Incorrect filter parameter")
      session['dqm.filter'] = filter

    if search != None:
      if not isinstance(search, str):
        raise HTTPError(500, "Incorrect search parameter")
      session['dqm.search'] = search

    if size != None:
      if not isinstance(size, str) \
         or size not in ("XS", "S", "M", "L", "XL"):
        raise HTTPError(500, "Incorrect size parameter")
      session['dqm.size'] = size

    if perpage != None:
      if not isinstance(perpage, str):
        raise HTTPError(500, "Incorrect per page parameter")
      elif perpage == "":
        session['dqm.perpage'] = self.sessiondef['dqm.perpage']
      elif not re.match(r"^\d+$", perpage):
        raise HTTPError(500, "Incorrect per page parameter")
      else:
        perpage = int(perpage)
        if perpage == 0:
          perpage = self.sessiondef['dqm.perpage']
        elif perpage < 1:
          perpage = 1
        elif perpage > 99:
          perpage = 99
        session['dqm.perpage'] = perpage

    if playmode != None:
      if not isinstance(playmode, str) \
         or playmode not in ("yes", "no"):
        raise HTTPError(500, "Incorrect play mode parameter")
      session['dqm.playmode'] = ((playmode == "yes" and 1) or 0)

    if playinterval != None:
      if not isinstance(playinterval, str) \
         or not re.match(r"^\d+$", playinterval):
        raise HTTPError(500, "Incorrect play interval parameter")
      playinterval = int(playinterval)
      if playinterval == 0:
        playinterval = self.sessiondef['dqm.playinterval']
      elif playinterval < 2:
        playinterval = 2
      elif playinterval > 99:
        playinterval = 99
      session['dqm.playinterval'] = playinterval

    if playpos != None:
      if not isinstance(playpos, str) \
         or not re.match(r"^\d+$", playpos):
        raise HTTPError(500, "Incorrect play position parameter")
      session['dqm.playpos'] = int(playpos)

    if root != None:
      if not isinstance(root, str):
        raise HTTPError(500, "Incorrect root parameter")
      session['dqm.root'][self.name] = self._checkObjectName(root)

    if picked != None:
      if not isinstance(picked, str):
        raise HTTPError(500, "Incorrect pick parameter")
      elif picked == "":
        if self.name in session['dqm.picked']:
          del session['dqm.picked'][self.name]
      else:
        session['dqm.picked'][self.name] = picked

    if focus != None:
      if not isinstance(focus, str):
        raise HTTPError(500, "Incorrect focus parameter")
      elif focus == "":
        if self.name in session['dqm.focus']:
          del session['dqm.focus'][self.name]
      else:
        session['dqm.focus'][self.name] = focus

    if page != None:
      if page == "":
        session['dqm.page'][self.name] = 1
      elif not re.match(r"^\d+$", page):
        raise HTTPError(500, "Incorrect page parameter")
      else:
        session['dqm.page'][self.name] = int(page)

    if reset == "yes":
      session['dqm.myobjs'][self.name] = []

    if add != None:
      session['dqm.myobjs'][self.name].extend \
        ([self._checkObjectName(x)
          for x in ((type(add) == str and [ add ]) or add)])

  # -----------------------------------------------------------------
  # Set dataset selection.
  def sessionSetQualityPlot(self, session, *args, **kwargs):
    name = kwargs.get("n", "")
    if not re.match(r"^[A-Za-z.\d]*$", name):
      raise HTTPError(500, "Invalid quality plot name")
    session['dqm.qplot'] = name
    self.gui._saveSession(session)
    return self._state(session)

  # Set dataset selection.
  def sessionSetDataset(self, session, *args, **kwargs):
    name = kwargs.get("n", "")
    session['dqm.dataset'] = name
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Set data step selection.
  def sessionSetStep(self, session, *args, **kwargs):
    name = kwargs.get("n", "")
    session['dqm.step'] = name
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Set run number selection.
  def sessionSetRun(self, session, *args, **kwargs):
    runnr = kwargs.get("n", None)
    if runnr == None or runnr == "Live" or runnr == "":
      session['dqm.runnr'] = "Live"
    elif not re.match(r"^\d+$", runnr):
      raise HTTPError(500, "Invalid run selection")
    else:
      session['dqm.runnr'] = runnr
      session['dqm.playmode'] = 0
      session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Set all/alarm/non-alarm filter.
  def sessionSetFilter(self, session, *args, **kwargs):
    name = kwargs.get("n", "all")
    if name not in ("all", "alarms", "nonalarms"):
      raise HTTPError(500, "Invalid filter setting")
    session['dqm.filter'] = name
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Set content search pattern.
  def sessionSetSearch(self, session, *args, **kwargs):
    session['dqm.search'] = kwargs.get("rx", "")
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Set the size of canvas objects.
  def sessionSetSize(self, session, *args, **kwargs):
    size = kwargs.get("sz", None)
    if size not in ("XS", "S", "M", "L", "XL"):
      raise HTTPError(500, "Incorrect size parameter")
    session['dqm.size'] = size
    self.gui._saveSession(session)
    return self._state(session)

  # Set the number of canvas objects per page.
  def sessionSetPerPage(self, session, *args, **kwargs):
    perpage = kwargs.get("n", "")
    if perpage == "":
      session['dqm.perpage'] = self.sessiondef['dqm.perpage']
    elif not re.match(r"^\d+$", perpage):
      raise HTTPError(500, "Incorrect per page parameter")
    else:
      perpage = int(perpage)
      if perpage == 0:
        perpage = self.sessiondef['dqm.perpage']
      elif perpage < 1:
        perpage = 1;
      elif perpage > 99:
        perpage = 99
      session['dqm.perpage'] = perpage
    self.gui._saveSession(session)
    return self._state(session)

  # Turn play mode on or off.
  def sessionSetPlayMode(self, session, *args, **kwargs):
    mode = kwargs.get("mode", "")
    if mode == "":
      session['dqm.playmode'] = self.sessiondef['dqm.playmode']
    elif mode == "yes":
      session['dqm.playmode'] = 1
    elif mode == "no":
      session['dqm.playmode'] = 0
    else:
      raise HTTPError(500, "Incorrect play mode parameter")
    self.gui._saveSession(session)
    return self._state(session)

  # Change play mode speed.
  def sessionSetPlayInterval(self, session, *args, **kwargs):
    playinterval = kwargs.get("v", "")
    if playinterval == "":
      session['dqm.playinterval'] = self.sessiondef['dqm.playinterval']
    elif not re.match(r"^\d+$", playinterval):
      raise HTTPError(500, "Incorrect play speed parameter")
    else:
      playinterval = int(playinterval)
      if playinterval == 0:
        playinterval = self.sessiondef['dqm.playinterval']
      elif playinterval < 2:
        playinterval = 2
      elif playinterval > 99:
        playinterval = 99
      session['dqm.playinterval'] = playinterval
    self.gui._saveSession(session)
    return self._state(session)

  # Change play head position in play mode.
  def sessionMovePlayPosition(self, session, *args, **kwargs):
    pos = kwargs.get("n", "")
    move = kwargs.get("move", "")
    if pos == "":
      session['dqm.playpos'] = self.sessiondef['dqm.playpos']
    elif not re.match(r"^\d+$", pos):
      raise HTTPError(500, "Incorrect play position parameter")
    else:
      session['dqm.playpos'] = int(pos)
    self.gui._saveSession(session)
    if move == "yes":
      return self._state(session)
    else:
      return "Ta."

  # Change the root shown in the selector menu.
  def sessionSetRoot(self, session, *args, **kwargs):
    name = kwargs.get("n", "")
    session['dqm.root'][self.name] = self._checkObjectName(name)
    if self.name in session['dqm.picked']:
      del session['dqm.picked'][self.name]
    if self.name in session['dqm.focus']:
      del session['dqm.focus'][self.name]
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Change the selected object.
  def sessionSetPicked(self, session, *args, **kwargs):
    name = kwargs.get("n", None)
    if name == None:
      if self.name in session['dqm.picked']:
        del session['dqm.picked'][self.name]
    elif not isinstance(name, str):
      raise HTTPError(500, "Incorrect pick operation")
    else:
      session['dqm.picked'][self.name] = self._checkObjectName(name)
    if self.name in session['dqm.focus']:
      del session['dqm.focus'][self.name]
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Change the selected object.
  def sessionSetFocus(self, session, *args, **kwargs):
    name = kwargs.get("n", None)
    if name == None:
      if self.name in session['dqm.focus']:
        del session['dqm.focus'][self.name]
    elif not isinstance(name, str):
      raise HTTPError(500, "Incorrect focus operation")
    else:
      session['dqm.focus'][self.name] = self._checkObjectName(name)
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

  # Change pages.
  def sessionSetPage(self, session, *args, **kwargs):
    i = int(kwargs.get('page', 1))
    session['dqm.page'][self.name] = i
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    self.gui._saveSession(session)
    return self._state(session)

# --------------------------------------------------------------------
class DQMSummaryWorkspace(Accelerator.DQMSummaryWorkspace, DQMWorkspace):
  def __init__(self, gui, rank, category, name):
    Accelerator.DQMSummaryWorkspace.__init__(self, gui, name)
    DQMWorkspace.__init__(self, gui, rank, category, name)

  # -----------------------------------------------------------------
  # Return JSON object for the current session state.  This method is
  # invoked when the GUI session page is loaded, which occurs either
  # when a completely new session has been started, the "reload"
  # button is pushed or the session URL was copied between browser
  # windows, or the session wants a period refresh.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

# --------------------------------------------------------------------
class DQMQualityWorkspace(Accelerator.DQMQualityWorkspace, DQMWorkspace):
  def __init__(self, gui, rank, category, name):
    Accelerator.DQMQualityWorkspace.__init__(self, gui, name)
    DQMWorkspace.__init__(self, gui, rank, category, name)

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

# --------------------------------------------------------------------
class DQMContentWorkspace(Accelerator.DQMContentWorkspace, DQMWorkspace):
  def __init__(self, gui, rank, category, name, match, *content):
    Accelerator.DQMContentWorkspace.__init__(self, gui, name, match)
    DQMWorkspace.__init__(self, gui, rank, category, name)
    self.content = content

  # Initialise a new session.
  def initialiseSession(self, session):
    DQMWorkspace.initialiseSession(self, session)
    if self.name not in session['dqm.myobjs'] and len(self.content):
      session['dqm.myobjs'][self.name] = deepcopy(self.content)
      session['dqm.picked'][self.name] = "Quick collection"

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

  # -----------------------------------------------------------------
  # Reset the canvas back to the workspace default.
  def sessionResetWorkspace(self, session, *args, **kwargs):
    session['dqm.playmode'] = 0
    session['dqm.playpos'] = 0
    session['dqm.size'] = self.sessiondef['dqm.size']
    session['dqm.perpage'] = self.sessiondef['dqm.perpage']
    for key in ('dqm.picked', 'dqm.focus', 'dqm.drawopts', 'dqm.myobjs'):
      if self.name in session[key]:
        del session[key][self.name]
    self.initialiseSession(session)
    self.gui._saveSession(session)
    return self._state(session)

  # -----------------------------------------------------------------
  # Change image parameters.
  def sessionAlterImage(self, session, *args, **kwargs):
    obj = kwargs['o']
    opt = kwargs['opt']
    val = kwargs['value']
    if not isinstance(obj, str):
      raise HTTPError(500, "Image operation without an image")
    if opt == "withref":
      if val == "":
        val == "no"
      if val not in ("yes", "no"):
        raise HTTPError(500, "Unrecognised reference setting")
    elif opt == "drawopts":
      if not re.match(r"^[A-Za-z,]*$", val):
        raise HTTPError(500, "Unknown draw options")
    elif opt in ("xtype", "ytype", "ztype"):
      if val == "":
	val = "lin"
      if val not in ("lin", "log"):
        raise HTTPError(500, "Unrecognised axis type")
    elif opt in ("xmin", "xmax", "ymin", "ymax", "zmin", "zmax"):
      if not re.match(r"^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)?$", val):
        raise HTTPError(500, "Unrecognised axis limit")
    else:
      raise HTTPError(500, "Attempt to set unknown image option")

    opts = session['dqm.drawopts'].setdefault(self.name, {}).setdefault(obj, {})
    if val == '':
      if opt in opts:
        del opts[opt]
      if len(opts) == 0:
        del session['dqm.drawopts'][self.name][obj]
      if len(session['dqm.drawopts'][self.name]) == 0:
        del session['dqm.drawopts'][self.name]
    else:
      opts[opt] = val

    self.gui._saveSession(session)
    return self._state(session)

# --------------------------------------------------------------------
class DQMShiftWorkspace(DQMContentWorkspace):
  SHIFT_ROOT = "00 Shift"

  def __init__(self, gui, rank, category, name):
    DQMContentWorkspace.__init__(self, gui, rank, category, name, "^")

  # Initialise a new session.
  def initialiseSession(self, session):
    DQMContentWorkspace.initialiseSession(self, session)
    session['dqm.root'][self.name] = self.SHIFT_ROOT

  # Override session state to fix root.
  def sessionState(self, session, *args, **kwargs):
    if not session['dqm.root'].get(self.name, "").startswith(self.SHIFT_ROOT):
      session['dqm.root'][self.name] = self.SHIFT_ROOT
      self.gui._saveSession(session)
    return DQMContentWorkspace.sessionState(self, session)

  # Change the root shown in the selector menu.
  def sessionSetRoot(self, session, *args, **kwargs):
    name = kwargs.get("n", "")
    if name.startswith("00 Shift"):
      return DQMContentWorkspace.sessionSetRoot(self, session, *args, **kwargs)
    else:
      return self._state(session)

  # Change the selected object.
  def sessionSetPicked(self, session, *args, **kwargs):
    name = kwargs.get("n", "")
    if name.startswith("00 Shift"):
      return DQMContentWorkspace.sessionSetPicked(self, session, *args, **kwargs)
    else:
      return self._state(session)

  # Change the selected object.
  def sessionSetFocus(self, session, *args, **kwargs):
    name = kwargs.get("n", None)
    if name == None or name.startswith("00 Shift"):
      return DQMContentWorkspace.sessionSetFocus(self, session, *args, **kwargs)
    else:
      return self._state(session)
